import numpy as np
from scipy import linalg
from itertools import chain
import os
from ..probabilistic_mapping_helpers import compute_X_hat, dist_pairwise_group

def compute_reconstruction_loss(x_orig, x_hat):
    """Computes the reconstruction loss (L_x)."""
    return np.mean((x_hat - x_orig) ** 2)

def compute_Ligf_group(x_orig, x_hat):
    """Computes the in-group-fairness loss for a single group."""
    D_X_orig = dist_pairwise_group(x_orig, x_orig, np.ones(x_orig.shape[1]))
    D_X_hat = dist_pairwise_group(x_hat, x_hat, np.ones(x_hat.shape[1]))
    return linalg.norm(D_X_orig - D_X_hat)


def compute_Ligf(X, X_hat, sensitive_groups, sensitive_indexes):
    """
    Computes the in-group-fairness (IGF) loss.
    """

    # Combine privileged and unprivileged groups into a single set
    groups = set(chain.from_iterable(
        list(sensitive_groups["privileged_groups"].values()) +
        list(sensitive_groups["unprivileged_groups"].values())
    ))

    # Identify non-sensitive feature indexes
    non_sensitive_indexes = [i for i in range(X.shape[1]) if i not in sensitive_indexes]

    L = []
    for group in groups:
        for s_index in sensitive_indexes:
            # Get mask for rows belonging to the current group
            mask = np.isin(X[:, s_index], group)

            if np.any(mask):  # Proceed only if group members exist
                x_orig = X[mask][:, non_sensitive_indexes]
                x_hat = X_hat[mask][:, non_sensitive_indexes]
                L.append(compute_Ligf_group(x_orig, x_hat))
    return np.sum(L)

def compute_gFair_fairness_loss_pairwise(D_X_f_s1_s2, D_X_f_hat_s1_s2):
    """Computes the group fairness loss based on distances between two groups."""
    group_gap = linalg.norm(D_X_f_s1_s2 - D_X_f_hat_s1_s2)
    return group_gap


def compute_gFair_group_fairness_loss(X, X_hat, params, sensitive_groups, sensitive_indexes, group_weights, D_X_f, biggest_gap=False):
    """Computes the group fairness loss based on distances between the groups."""
    L_z = 0
    N, M = X.shape
    alpha = params[:M]
    if biggest_gap:
        max_group_avg_difference = 0
    for s_attr in sensitive_groups["privileged_groups"].keys():
        privileged_groups = sensitive_groups["privileged_groups"][s_attr]
        unprivileged_groups = sensitive_groups["unprivileged_groups"][s_attr]
        for index_advantaged_group in range(len(privileged_groups)):
            for s_index in sensitive_indexes:
                mask_advantaged = np.isin(element=X[:, s_index],
                                          test_elements=privileged_groups[index_advantaged_group]).reshape(-1)
                if sum(mask_advantaged) != 0:
                    for index_sensitive_group in range(index_advantaged_group + 1, len(unprivileged_groups)):
                        if unprivileged_groups[index_sensitive_group] != privileged_groups[index_advantaged_group]:
                            mask_s = np.isin(element=X[:, s_index],
                                             test_elements=unprivileged_groups[index_sensitive_group]).reshape(-1)
                            if sum(mask_s) != 0:
                                if unprivileged_groups[index_sensitive_group] != privileged_groups[
                                    index_advantaged_group]:
                                    weight = abs(
                                        group_weights[privileged_groups[index_advantaged_group]] - group_weights[
                                            unprivileged_groups[index_sensitive_group]]) + 1
                                    D_X_f_hat_s1_s2 = dist_pairwise_group(X_hat[mask_s], X_hat[mask_advantaged],
                                                                          alpha)
                                    D_X_f_s1_s2 = D_X_f[mask_s][:, mask_advantaged]
                                    if biggest_gap:
                                        if max_group_avg_difference < compute_gFair_fairness_loss_pairwise(D_X_f_s1_s2, D_X_f_hat_s1_s2):
                                            max_group_avg_difference = compute_gFair_fairness_loss_pairwise(D_X_f_s1_s2, D_X_f_hat_s1_s2)
                                            L_z = [weight * compute_gFair_fairness_loss_pairwise(D_X_f_s1_s2, D_X_f_hat_s1_s2)]
                                    else:
                                        L_z += weight * compute_gFair_fairness_loss_pairwise(D_X_f_s1_s2, D_X_f_hat_s1_s2)

    return L_z


def gFair_optimisation(params, X, sensitive_groups, sensitive_indexes, D_X_f, group_weights, biggest_gap, logs_path='', k=10, A_x=1e-4, A_z=1e-4, A_igf=1e-4):
    """Optimisation function for gFair."""
    gFair_optimisation.iters += 1
    X_hat, _ = compute_X_hat(X, params, k, alpha=True)

    L_x = compute_reconstruction_loss(X, X_hat)
    L_igf = compute_Ligf(X, X_hat, sensitive_groups, sensitive_indexes)
    L_z = compute_gFair_group_fairness_loss(X, X_hat, params, sensitive_groups, sensitive_indexes, group_weights, D_X_f,
                                            biggest_gap)

    criterion = A_x * L_x + A_z * L_z + A_igf * L_igf
    if gFair_optimisation.iters % 100 == 0:
        print("step: {}, L_x: {},  L_z: {},  L_igf: {}, loss:{}\n".format(
                    gFair_optimisation.iters, L_x, L_z, L_igf, criterion))
        if logs_path != '':
            os.makedirs(logs_path, exist_ok=True)
            with open(os.path.join(logs_path, "logs.txt"), 'a') as f:
                f.write("step: {}, L_x: {},  L_z: {},  L_igf: {}, loss:{}\n".format(
                    gFair_optimisation.iters, L_x, L_z, L_igf, criterion))

    return criterion



from transformers import AutoModelForCausalLM, AutoTokenizer
from tqdm import tqdm



class LLM_caller(object):
    def __init__(self, config):
        self.config = config
        self.llm_type = config['llm_type']
        self.device = config['device']
        if self.llm_type == 'api':
            self.init_llm_api(config['llm_name'], config['api_key'], config['api_base'],
                              config['max_tokens'], config['temperature'])
        elif self.llm_type == 'local':
            self.init_llm_local(config['llm_name'], config['llm_path_dict'],
                                config['max_tokens'], config['use_8bit'], config['device_map'])
        elif self.llm_type == 'vllm':
            self.init_llm_vllm(config['llm_name'], config['llm_path_dict'],
                                config['temperature'], config['max_tokens'])
        else:
            raise ValueError('llm type not found.')

    def clear(self):
        if self.llm_type == 'vllm':
            del self.vllm
            del self.tokenizer
        if self.llm_type =='local':
            del self.model
            del self.tokenizer


    def init_llm_api(self, llm_name, api_key, api_base='EMPTY', max_tokens=256, temperature=0.8):
        print(f'use api llm for generating...')
        self.llm_name = llm_name
        self.api_key = api_key
        self.api_base = api_base
        self.max_tokens = max_tokens
        self.temperature = temperature
        self.llm_func = self.GetResultFromGPT


    def init_llm_local(self, llm_name, llm_path_dict, max_tokens=512, use_8bit=False, device_map='auto'):
        """
        Initialize a local language model (LLM) for text generation.

        This method sets up a local instance of a causal language model using Hugging Face's transformers library. It loads the model and tokenizer from the provided path, with options for controlling memory usage and device allocation.


        :param llm_name: `str`
            The name of the language model to initialize. Must correspond to a key in the `llm_path_dict`.
        :param llm_path_dict: `Dict[str, str]`
            A dictionary mapping model names to their local directory paths.
        :param max_tokens: `int`, optional
            The maximum number of tokens to generate. Defaults to 512.
        :param use_8bit: `bool`, optional
            Whether to use 8-bit quantization for loading the model, reducing memory usage at the cost of speed. Defaults to `False`.
        :param device_map: `str`, optional
            Specifies how to allocate the model across devices. 'auto' automatically balances the model on available devices. Custom mappings can also be defined. Defaults to 'auto'.


        """
        print(f'use local llm for generating...')
        assert llm_name in llm_path_dict.keys(), f"LLM {llm_name} Path Not Found"
        self.llm_name = llm_name
        self.llm_path = llm_path_dict[llm_name]
        self.max_tokens = max_tokens
        self.model = AutoModelForCausalLM.from_pretrained(
            self.llm_path,
            load_in_8bit=use_8bit,
            torch_dtype="auto",
            device_map=device_map,
            trust_remote_code=True
        )
        self.tokenizer = AutoTokenizer.from_pretrained(self.llm_path, trust_remote_code=True)
        self.llm_func = self.GetResultFromHuggingFace

    def init_llm_vllm(self, llm_name, llm_path_dict, temperature=0.8, max_tokens=256):
        """
        Initialize the Language Model using vLLM for text generation.

        This method sets up the vLLM framework to utilize a specified language model for generating text. It configures
        sampling parameters like temperature and maximum tokens, selects the tokenizer corresponding to the language model,
        and determines whether to use batch processing based on the provided configuration.

        :param llm_name: (str) The name of the language model to be initialized.
        :param llm_path_dict: (dict) A dictionary mapping language model names to their respective local paths.
        :param temperature: (float, optional) The sampling temperature for generated text. Defaults to 0.8.
        :param max_tokens: (int, optional) The maximum number of tokens to generate. Defaults to 256.


        """
        print(f'use vllm for generating...')
        from vllm import LLM, SamplingParams
        self.use_batch = self.config['use_batch']
        self.llm_path = llm_path_dict[llm_name]
        self.tokenizer = AutoTokenizer.from_pretrained(self.llm_path, trust_remote_code=True)
        self.llm_name = llm_name
        self.vllm = LLM(model=self.llm_path, max_model_len=8192)
        self.sampling_params = SamplingParams(temperature=temperature, max_tokens=max_tokens)
        self.llm_func = self.GetBatchResultFromVllm if self.use_batch else self.GetResultFromVllm

    def GetBatchResultFromVllm(self, data):
        """
        This method processes a batch of data to generate text using the VLLM engine, with customization based on the LLM name.

        :param data : List[Dict[str, str]]
            A list of dictionaries, each containing 'input' and 'instruction' keys for generating text.

        :return: List[str]
            A list of generated texts corresponding to the input data.
        """
        if 'Mistral' in self.llm_name:
            messages_list = [[
                {"role": "user", "content": prompt_dict['input'] + '\n'+ prompt_dict['instruction']},
            ] for prompt_dict in data]
        else:
            messages_list = [[
                {"role": "system", "content": prompt_dict['input']},
                {"role": "user", "content": prompt_dict['instruction']},
            ] for prompt_dict in data]
        prompt_ids = [self.tokenizer.apply_chat_template(messages, add_generation_prompt=True) for messages in messages_list ]
        outputs = self.vllm.generate(prompt_token_ids=prompt_ids,
                           sampling_params=self.sampling_params,
                            use_tqdm= False
                           )
        generated_text = [output.outputs[0].text for output in outputs]
        return generated_text


    def GetResultFromVllm(self, prompt_dict):
        """
        Obtains a generated text result from the VLLM model given a dictionary containing input and instruction data.

        :param: prompt_dict (dict): A dictionary carrying two keys:
            - 'input' (str): The context or initial information for the conversation.
            - 'instruction' (str): The specific directive or query for the model to act upon.

        :return: str: The generated text output from the VLLM model corresponding to the provided inputs.

        Note:
        - Ensure `self.tokenizer` and `self.vllm` are properly initialized and configured before calling this method.
        - The `sampling_params` should be set to appropriate values within the class instance for desired generation behavior.
        """
        messages = [
            {"role": "system", "content": prompt_dict['input']},
            {"role": "user", "content": prompt_dict['instruction']},
        ]
        prompt_ids = self.tokenizer.apply_chat_template(messages, add_generation_prompt=True)
        outputs = self.vllm.generate(prompt_token_ids=[prompt_ids],
                           sampling_params=self.sampling_params,
                            use_tqdm= False
                           )
        print(outputs[0].outputs[0].text)
        return outputs[0].outputs[0].text

    def GetResultFromHuggingFace(self, prompt_dict):
        """
        GetResultFromHuggingFace(self, prompt_dict: dict) -> str

        Generates a response from a Hugging Face language model based on the provided prompt dictionary.

        This method constructs a message context using the 'input' and 'instruction' keys from the `prompt_dict`.
        It applies a chat template to format the messages, tokenizes them, and prepares the input for the model.
        The model's generate function is then used to produce new text, which is decoded and returned as the final response.

        :param **prompt_dict** (dict): A dictionary containing two keys:
        :param 'input' (str): The initial context or scenario for the conversation.
        :param 'instruction' (str): The user's instruction or query within the given context.

        :return: **response** (str): The generated text output by the language model in response to the prompt.
        """
        messages = [
            {"role": "system", "content": prompt_dict['input']},
            {"role": "user", "content": prompt_dict['instruction']},
        ]

        text = self.tokenizer.apply_chat_template(
            messages,
            tokenize=False,
            return_tensors="pt",
            add_generation_prompt=True,
            # return_dict=True
        )

        model_inputs = self.tokenizer([text], return_tensors="pt").to(self.device)  # if args.model != 'glm' else text.to('cuda')
        # else:
        # model_inputs = tokenizer.build_inputs_for_generation(text, allowed_special="all", return_tensors="pt",padding=True).to('cuda')
        generated_ids = self.model.generate(
            model_inputs.input_ids,
            max_new_tokens=self.max_tokens,
            pad_token_id= self.tokenizer.eos_token_id
        )

        generated_ids = [
            output_ids[len(input_ids):] for input_ids, output_ids in zip(model_inputs.input_ids, generated_ids)
        ]

        response = self.tokenizer.batch_decode(generated_ids, skip_special_tokens=True)[0]
        # print(f'response:{response}')
        return response

    def GetResultFromGPT(self, prompt_dict):
        """
        GetResultFromGPT(self, prompt_dict: dict) -> str

        Sends a request to an OpenAI-compatible API server, specifically vLLM, using the provided prompt dictionary to generate a text completion based on a pre-configured language model.

        This method constructs a chat completion request with a system message and a user message extracted from the `prompt_dict`. It then sends the request to the API endpoint defined by `openai_api_base`, utilizing the API key `openai_api_key`. The response content from the API is returned as a string.

        :param prompt_dict (dict): A dictionary containing two keys:
        :param 'input': A string representing the system's input or context for the conversation.
        :param 'instruction': A string representing the user's message or instruction for generating a response.

        :return: str: The generated text content from the API's response.

        Note:
            - Ensure that `self.api_key` and `self.api_base` are set appropriately before calling this method.
            - The model name (`self.llm_name`), maximum number of tokens to generate (`self.max_tokens`), and temperature setting (`self.temperature`) are used to configure the generation parameters and must be predefined within the instance.

        """
        from openai import OpenAI
        # Set OpenAI's API key and API base to use vLLM's API server.
        openai_api_key = self.api_key
        openai_api_base = self.api_base
        client = OpenAI(
            api_key=openai_api_key,
            base_url=openai_api_base,
        )

        chat_response = client.chat.completions.create(
            model=self.llm_name,
            messages=[
                {"role": "system", "content": prompt_dict['input']},
                {"role": "user", "content": prompt_dict['instruction']},
            ],
            max_tokens=self.max_tokens,
            temperature=self.temperature,

        )
        response = chat_response.choices[0].message.content
        # print(response)
        return response

    def batch_list(self, input_list, batch_size):
        """
        Yields batches of a list in specified size.

        This method takes an input list and divides it into smaller sublists (batches),
        each of which has a maximum length defined by the `batch_size` parameter. It yields
        these batches one at a time, allowing for efficient processing of large lists in chunks.

        :param input_list : list
            The list to be divided into batches.
        :param batch_size : int
            The maximum size of each batch. Must be a positive integer.

        :return: list
            Sublists (batches) of the input list, each with up to `batch_size` elements.

        """
        for i in range(0, len(input_list), batch_size):
            yield input_list[i:i + batch_size]

    def get_response(self, data):
        """
        Get the response from the language model based on the input data.

        This method processes the input data and retrieves responses from the
        configured language model. It supports both single-input and batched
        processing modes, depending on the configuration of the object.

        In the case of a 'vllm' type model with batch processing enabled,
        the input data is divided into batches according to the specified batch size.
        Each batch is processed sequentially, with the results being collected
        and merged back with the original input data structure. For other models or
        when batch processing is not used, each input item is processed individually.

        :param data (List[Dict]): A list of dictionaries, where each dictionary contains
                               the necessary input for the language model's `llm_func`.

        :return: List[Dict]: A list of dictionaries, where each dictionary from the input `data`
                        is extended with a new key `'predict'` holding the respective response
                        from the language model.

        Note:
            The method currently prints progress bars using `tqdm` for visual feedback
            during batch processing. The commented out code at the end suggests a file
            saving operation which is not executed in this function's scope.
        """
        json_list = []
        print(f'-------- Get Model {self.llm_name} Response-------')

        if self.llm_type == 'vllm' and self.use_batch:
            batched_data = list(self.batch_list(data, self.config['batch_size']))
            all_result = []
            for batch in tqdm(batched_data):
                batched_result = self.llm_func(batch)
                all_result.extend(batched_result)
            for prompt_dict, r in zip(data, all_result):
                prompt_dict['predict'] = r
                json_list.append(prompt_dict)
        else:
            for prompt_dict in tqdm(data):
                response = self.llm_func(prompt_dict)
                # print(f'response:{response}')
                prompt_dict['predict'] = response
                json_list.append(prompt_dict)

        return json_list



import os

import pandas as pd
import yaml

from .datasets.COMPAS import Compas
from .preprocessing_model import fairness_method_mapping
from .ranker_model import ranker_mapping


class RankerTrainer(object):
    def __init__(self, train_config):
        self.train_config = train_config
        self.load_configs()

    def load_configs(self):
        # Load the configs for the dataset and for the pre-processing method
        with open(os.path.join(self.train_config['task'], "properties", "dataset",
                               self.train_config['dataset'].lower() + ".yaml"), 'r') as f:
            data_config = yaml.safe_load(f)
        self.train_config["data_reader_class"] = data_config["data_reader_class"]

        print(self.train_config['train_ranker_config']['preprocessing_model'])
        with open(os.path.join(self.train_config['task'], "properties", "models",
                               self.train_config['train_ranker_config']['preprocessing_model'] + ".yaml"),
                  'r') as f:
            model_config = yaml.safe_load(f)
        self.train_config["pre_processing_config"] = model_config["pre_processing_config"]

        print(self.train_config)

    def train(self):
        # train the ranker with the pre-processing method
        if self.train_config["data_reader_class"]["name"] == "COMPAS":
            dataset = Compas(self.train_config["data_reader_class"])

        if "pre_processing_config" in self.train_config:
            fairness_method = fairness_method_mapping[self.train_config["pre_processing_config"]["name"]](
                self.train_config["pre_processing_config"], dataset)

        ranker = ranker_mapping[self.train_config['train_ranker_config']['name']](
            self.train_config['train_ranker_config'], dataset)

        for run in range(dataset.k_fold):
            data_train, data_test = dataset.read_data(run)

            if "pre_processing_config" in self.train_config:
                if self.train_config["pre_processing_config"]["qid_process"]:
                    data_train_fair = []
                    data_test_fair = []
                    qids = data_train[dataset["query_col"]].unique()
                    for qid in qids:
                        root_model_path = fairness_method.model_path.copy()
                        fairness_method.model_path = os.path.join(root_model_path, str(run), str(qid))

                        data_qid_train = data_train[dataset["query_col"] == qid]
                        data_qid_test = data_test[dataset["query_col"] == qid]

                        fairness_method.fit(data_qid_train, str(run))

                        data_qid_train_fair = fairness_method.transform(data_qid_train, str(run))
                        data_qid_test_fair = fairness_method.transform(data_qid_test, str(run))

                        data_train_fair.append(data_qid_train_fair)
                        data_test_fair.append(data_qid_test_fair)

                    data_train_fair = pd.concat(data_train_fair)
                    data_train_fair.to_csv(os.path.join(fairness_method.fair_data_path, 'fair_train_data.csv'))

                    data_test_fair = pd.concat(data_test_fair)
                    data_test_fair.to_csv(os.path.join(fairness_method.fair_data_path, 'fair_test_data.csv'))
                else:

                    fairness_method.fit(data_train, str(run))
                    data_train_fair = fairness_method.transform(data_train, str(run), "train")
                    data_test_fair = fairness_method.transform(data_test, str(run), "test")

                ranker_train_data = data_train_fair
                ranker_test_data = data_test_fair
            else:
                ranker_train_data = data_train
                ranker_test_data = data_test
            ranker.train(ranker_train_data, ranker_test_data, run)
            prediction = ranker.predict(data_test_fair, run, "test")

            ranker.evaluate_run(prediction, run, "test")

        ranker.evaluate("test")  # computes mean over runs



import numpy as np
from .Abstract_Ranker import Abstract_Regularizer
import torch.nn.functional as F
import torch
import torch.nn as nn

r"""
DPR
################################################
@inproceedings{10.1145/3397271.3401177,
author = {Zhu, Ziwei and Wang, Jianling and Caverlee, James},
title = {Measuring and Mitigating Item Under-Recommendation Bias in Personalized Ranking Systems},
year = {2020},
isbn = {9781450380164},
publisher = {Association for Computing Machinery},
address = {New York, NY, USA},
url = {https://doi.org/10.1145/3397271.3401177},
doi = {10.1145/3397271.3401177},
booktitle = {Proceedings of the 43rd International ACM SIGIR Conference on Research and Development in Information Retrieval},
pages = {449–458},
numpages = {10},
keywords = {statistical parity, recommender systems, recommendation bias, equal opportunity},
location = {Virtual Event, China},
series = {SIGIR '20}
}

##Note that in this code, we extend the two-group cases into multi-group cases

"""

class DPR(Abstract_Regularizer):
    def __init__(self, config, group_weight):
        super().__init__(config)

        self.mlp = nn.Sequential(
            nn.Linear(self.config['embedding_size'], self.config['hidden_size']),  # 第一层
            nn.ReLU(),  # 第一层的激活函数
            nn.Linear(self.config['hidden_size'], self.config['group_num'])  # 第二层
        )

        self.group_loss = nn.CrossEntropyLoss()
        self.type = ["pair"]


    def fairness_loss(self, input_dict, Model):
        pos_items = input_dict['item_ids']
        neg_items = input_dict['neg_item_ids']

        pos_groups = input_dict['group_ids']
        neg_groups = input_dict['neg_group_ids']

        scores = input_dict['scores'] #[B]

        pos_item_embs = Model.item_embedding(pos_items)
        neg_item_embs = Model.item_embedding(neg_items)

        pos_group_predict = self.mlp(pos_item_embs)
        neg_group_predict = self.mlp(neg_item_embs)

        loss_pos_group = self.group_loss(pos_group_predict, pos_groups)
        loss_neg_group = self.group_loss(neg_group_predict, neg_groups)

        #group_scores =
        Q = torch.full_like(scores, 1.0 / self.config['item_num'])
        loss_uniform = scores * (torch.log(scores) - torch.log(Q))
        loss_uniform = torch.mean(loss_uniform)

        fairloss = -self.config['alpha'] * (loss_pos_group + loss_neg_group) + self.config['beta'] * loss_uniform

        return fairloss










import os
import pickle
import torch
import numpy as np
from tqdm import tqdm
from sklearn.model_selection import KFold
from ..utils.utils import load_embedding, read_rel_feat
from ..utils.div_type import div_dataset


def gen_data_file_train(train_qids, qd, train_data, doc_emb, query_emb, rel_feat, save_path):
    """
    Generates and saves training data files with document embeddings, query suggestions, and relevance features.
    This function processes training query IDs to create structured data samples containing document vectors,
    query suggestion vectors, relevance features, and various masking tensors for handling variable-length inputs.
    
    :param train_qids: List of training query IDs to process
    :param qd: Dictionary containing query-document information
    :param train_data: Dictionary containing training data pairs
    :param doc_emb: Dictionary of document embeddings
    :param query_emb: Dictionary of query embeddings
    :param rel_feat: Dictionary of relevance features
    :param save_path: Path where the processed training data will be saved
    """
    data_list = []  # {qid:, query:, doclist:, doc2vec:, sub2vec:, rel_feat:, sub_rel_feat:, list_pair:}
    #max_d, max_s, max_ps, max_ns = 0, 0, 0, 0
    for qid in tqdm(train_qids):
        doc2vec = [doc_emb[docid] for docid in qd[qid].best_docs_rank]
        sub2vec = [query_emb[query_sugg] for query_sugg in qd[qid].query_suggestion]

        for i in range(len(train_data[qid])):
            '''
            max_d, max_s, max_ps, max_ns = max(max_d, len(doc2vec)), max(max_s, len(sub2vec)), \
                                           max(max_ps, len(temp['pos_subrel_feat'])), \
                                           max(max_ns, len(temp['neg_subrel_feat']))
            '''
            temp = {}
            temp['qid'] = qid
            temp['query'] = qd[qid].query
            temp['doclist'] = qd[qid].best_docs_rank
            temp['doc2vec_mask'] = torch.tensor([1]*len(doc2vec)+[0]*(50-len(doc2vec)))
            temp['sub2vec_mask'] = torch.tensor([1]*len(sub2vec)+[0]*(10-len(sub2vec)))
            temp['doc2vec'] = torch.tensor(doc2vec+[[0]*100]*(50-len(doc2vec)))
            temp['sub2vec'] = torch.tensor(sub2vec+[[0]*100]*(10-len(sub2vec)))
            temp['positive_mask'] = train_data[qid][i][1]
            temp['negative_mask'] = train_data[qid][i][2]
            temp['weight'] = train_data[qid][i][3]
            pos_id = qd[qid].best_docs_rank[int(torch.argmax(train_data[qid][i][1]))]
            neg_id = qd[qid].best_docs_rank[int(torch.argmax(train_data[qid][i][2]))]
            temp['pos_qrel_feat'] = torch.Tensor(rel_feat[qd[qid].query][pos_id])
            temp['neg_qrel_feat'] = torch.Tensor(rel_feat[qd[qid].query][neg_id])
            temp['subrel_feat_mask'] = torch.tensor([1]*len(qd[qid].query_suggestion)+[0]*(10-len(qd[qid].query_suggestion)))
            temp['pos_subrel_feat'] = torch.tensor([rel_feat[query_sugg][pos_id] for query_sugg in
                                                qd[qid].query_suggestion]+[[0]*18]*(10-len(qd[qid].query_suggestion)))
            temp['neg_subrel_feat'] = torch.tensor([rel_feat[query_sugg][neg_id] for query_sugg in
                                       qd[qid].query_suggestion]+[[0]*18]*(10-len(qd[qid].query_suggestion)))
            data_list.append(temp)
    torch.save(data_list, save_path)
    #return max_d, max_s, max_ps, max_ns


def gen_data_file_test(test_qids, qd, test_data, doc_emb, query_emb, rel_feat, save_path):
    """
    Generates and saves test data files with document embeddings, query suggestions, and relevance features.
    This function processes test query IDs to create a structured dictionary of test samples, including
    document vectors, query suggestion vectors, and relevance features with appropriate padding and masking.
    
    :param test_qids: List of test query IDs to process
    :param qd: Dictionary containing query-document information
    :param test_data: Dictionary containing test data pairs
    :param doc_emb: Dictionary of document embeddings
    :param query_emb: Dictionary of query embeddings
    :param rel_feat: Dictionary of relevance features
    :param save_path: Path where the processed test data will be saved
    """
    data_list = {}  # {qid:, query:, doclist:, doc2vec:, sub2vec:, rel_feat:, sub_rel_feat:, list_pair:}
    for qid in tqdm(test_qids):
        data_list[qid] = {}
        doc2vec = [doc_emb[docid] for docid in qd[qid].best_docs_rank]
        sub2vec = [query_emb[query_sugg] for query_sugg in qd[qid].query_suggestion]
        data_list[qid]['qid'] = qid
        data_list[qid]['query'] = qd[qid].query
        data_list[qid]['doclist'] = qd[qid].best_docs_rank
        data_list[qid]['doc2vec_mask'] = torch.tensor([1] * len(doc2vec) + [0] * (50 - len(doc2vec)))
        data_list[qid]['sub2vec_mask'] = torch.tensor([1] * len(sub2vec) + [0] * (10 - len(sub2vec)))
        data_list[qid]['doc2vec'] = torch.tensor(doc2vec + [[0] * 100] * (50 - len(doc2vec)))
        data_list[qid]['sub2vec'] = torch.tensor(sub2vec + [[0] * 100] * (10 - len(sub2vec)))
        data_list[qid]['pos_qrel_feat'] = torch.Tensor([rel_feat[qd[qid].query][pos_id]
                                    for pos_id in qd[qid].best_docs_rank]+[[0]*18]*(50-len(qd[qid].best_docs_rank)))
        pos_subrel_feat = [] # 50*10*18
        subrel_mask = []
        for pos_id in qd[qid].best_docs_rank:
            temp1 = []
            for query_sugg in qd[qid].query_suggestion:
                temp1.append(rel_feat[query_sugg][pos_id])
            temp2 = [0]*len(temp1)+[1]*(10-len(temp1))
            temp1.extend([[0]*18]*(10-len(temp1)))
            pos_subrel_feat.append(temp1)
            subrel_mask.append(temp2)
        subrel_mask.extend([[0] * 10] * (50 - len(pos_subrel_feat)))
        pos_subrel_feat.extend([[[0]*18]*10]*(50-len(pos_subrel_feat)))
        data_list[qid]['subrel_feat_mask'] = torch.tensor(subrel_mask)
        data_list[qid]['pos_subrel_feat'] = torch.tensor(pos_subrel_feat)
    torch.save(data_list, save_path)


def divide_five_fold_train_test(config):
    """
    Performs 5-fold cross-validation data splitting and preprocessing for model training and testing.
    This function loads necessary embeddings and features, splits the data into 5 folds, and generates
    processed training and test data files for each fold using the specified configuration settings.
    
    :param config: Dictionary containing configuration parameters including data directories, 
                  embedding types, and model settings
    """
    all_qids = np.load(os.path.join(config['data_dir'], 'all_qids.npy'))
    qd = pickle.load(open(os.path.join(config['data_dir'], 'div_query.data'), 'rb'))
    train_data = pickle.load(open(os.path.join(config['data_dir'], config['model'], 'listpair_train.data'), 'rb'))
    doc_emb = load_embedding(os.path.join(config['data_dir'], config['embedding_dir'], config['embedding_type']+'_doc.emb'))
    query_emb = load_embedding(os.path.join(config['data_dir'], config['embedding_dir'], config['embedding_type']+'_query.emb'))
    rel_feat = read_rel_feat(os.path.join(config['data_dir'], 'rel_feat.csv'))

    data_dir = os.path.join(config['data_dir'], config['model'], 'fold/')
    if not os.path.exists(data_dir):
        os.makedirs(data_dir)
    fold = 0
    for train_ids, test_ids in KFold(5).split(all_qids):
        fold += 1
        res_dir = os.path.join(data_dir, 'fold'+str(fold))
        if not os.path.exists(res_dir):
            os.makedirs(res_dir)
        train_ids.sort()
        test_ids.sort()
        train_qids = [str(all_qids[i]) for i in train_ids]
        test_qids = [str(all_qids[i]) for i in test_ids]
        '''{qid:, query:, doc2vec:, sub2vec:, rel_feat:, sub_rel_feat:, list_pair:} '''
        gen_data_file_train(train_qids, qd, train_data, doc_emb, query_emb, rel_feat, os.path.join(res_dir, 'train_data.pkl'))
        gen_data_file_test(test_qids, qd, train_data, doc_emb, query_emb, rel_feat, os.path.join(res_dir, 'test_data.pkl'))


def Process(config):
    """
    Initializes and executes the complete data processing pipeline for the model.
    This function creates a dataset object, generates list-pair training data, and performs
    5-fold cross-validation data splitting and preprocessing.
    
    :param config: Dictionary containing configuration parameters for data processing
    """
    D = div_dataset(config)
    D.get_listpair_train_data()
    divide_five_fold_train_test(config)


"""Learning fair representations is a pre-processing technique that finds a
    latent representation which encodes the data well but obfuscates information
    about protected attributes [2]_.
    References:
        .. [2] R. Zemel, Y. Wu, K. Swersky, T. Pitassi, and C. Dwork,  "Learning
           Fair Representations." International Conference on Machine Learning,
           2013.
    Based on code from https://github.com/zjelveh/learning-fair-representations
    """

import numpy as np
import pandas as pd
import scipy.optimize as optim
import os
from .modules.LFR.loss import LFR_optimisation as LFR_func
from .modules.probabilistic_mapping_helpers import compute_X_hat
from .utils import process_data_input, process_data_output, save_model_data, load_model_data
from .fair_model import PreprocessingFairnessIntervention

class LFR(PreprocessingFairnessIntervention):
    """
        Learning Fair Representations (LFR) fairness intervention.

        This class applies the LFR approach to modify the dataset such that fairness constraints
        are met while preserving as much utility as possible.
        """
    def __init__(self, configs, dataset):
        """
        Initialize the LFR model with the given configurations and dataset.

        :param configs : dict
            Configuration dictionary containing model parameters.
        :param dataset : str
            The dataset to be processed.
        """
        super().__init__(configs, dataset)
    def fit(self, X_train, run):
        """
        Train the LFR fairness model using the given training dataset.

        This method optimizes a fairness objective by learning fair representations
        of the data using constrained optimization.

        :param X_train : pandas.DataFrame or numpy.ndarray
            The training dataset. It is assumed that the last non-sensitive column is the target variable.
        :param run : str
            The identifier for the training run.

        :return : self
            The trained LFR model.
        """
        if not os.path.exists(os.path.join(self.model_path, run)):
            X_train, group_weights, sensitive_groups, sensitive_column_indices, nonsensitive_column_indices = (
                process_data_input(X_train, self.configs, self.dataset))

            if self.configs["seed"] is not None:
                np.random.seed(self.configs["seed"])

            # assumes that the last non-sensitive column of X_train is Y_train


            Y_train = X_train[:, nonsensitive_column_indices][-1]
            features_dim = X_train.shape[1]

           # Initialize the LFR_module optim objective parameters
            parameters_initialization = np.random.uniform(size=int(self.configs["k"] + features_dim * self.configs["k"]))

            bnd = [(0, 1)] * self.configs["k"] + [(None, None)] * features_dim * self.configs["k"]
            LFR_func.steps = 0


            self.opt_params = optim.fmin_l_bfgs_b(LFR_func, x0=parameters_initialization, epsilon=1e-5,
                                                          args=(X_train, Y_train, sensitive_groups, sensitive_column_indices, self.configs["k"],
                                                                self.configs["A_x"], self.configs["A_y"], self.configs["A_z"],
                                                                group_weights, self.configs["biggest_gap"],
                                                                os.path.join(self.model_path, run)),
                                                          bounds=bnd, approx_grad=True, maxfun=self.configs["maxfun"],
                                                          maxiter=self.configs["maxiter"], disp=False)[0]
            self.w = self.opt_params[:self.configs["k"]]
            self.prototypes = self.opt_params[self.configs["k"]:].reshape((self.configs["k"], features_dim))

            save_model_data(self, os.path.join(self.model_path, run))
        else:
            self.opt_params = load_model_data(os.path.join(self.model_path, run))
        return self

    def transform(self, X, run, file_name=None):
        """
            Apply the fairness transformation to the dataset using the learned model.

            This method ensures fairness by adjusting feature distributions while maintaining data utility.

            :param X : pandas.DataFrame
                The dataset to which the fairness transformation is applied.
            :param run : str
                The identifier for the transformation run.
            :param file_name : str, optional
                Name of the file to save the transformed dataset.

            :return : pandas.DataFrame
                The dataset with transformed fair columns.
        """
        fair_data_path = os.path.join(self.fair_data_path, run)
        os.makedirs(fair_data_path, exist_ok=True)
        fair_data_file = os.path.join(fair_data_path, f'fair_{file_name}_data.csv')
        if not os.path.exists(fair_data_file):
            X_np, group_weights, sensitive_groups, sensitive_column_indices, nonsensitive_column_indices = (
                process_data_input(X, self.configs, self.dataset))

            X_hat, _ = compute_X_hat(X_np, self.opt_params, self.configs["k"], alpha=False)
            X_fair = process_data_output(X, X_hat, self.dataset, nonsensitive_column_indices, fair_data_path, file_name)
        else:
            X_fair = pd.read_csv(fair_data_file)
        return X_fair





import numpy as np
import pandas as pd
import os
import random
import copy
import yaml
from sklearn.model_selection import train_test_split
import json
from tqdm import tqdm,trange

def build_map(df,col_name):
    """
        Encodes a categorical column in a DataFrame by mapping unique values to integers
        and returns a decoding dictionary.

        :param df: The input DataFrame.
        :param col_name: The name of the column to be encoded.
        :return: dict: A dictionary mapping encoded integers back to original values.
    """
    key = df[col_name].unique().tolist()
    encode = dict(zip(key, range(len(key))))
    decode = dict(zip(range(len(key)),key))
    df[col_name] = df[col_name].apply(lambda x: encode[x])
    return decode

def build_item_map(df,col_name, history_name):
    """
        Encodes a categorical column in a DataFrame by mapping unique values with historical records to integers
        and returns a decoding dictionary.

        :param df: The input DataFrame.
        :param col_name: The name of the column to be encoded.
        :return: dict: A dictionary mapping encoded integers back to original values.
    """


    item_set = []
    for i in range(len(df)):
        item_set.append(df[col_name][i])
        for h_i in df[history_name][i]:
            item_set.append(h_i)

    key = list(set(item_set))
    encode = dict(zip(key, range(len(key))))
    decode = dict(zip(range(len(key)),key))
    df[col_name] = df[col_name].apply(lambda x: encode[x])
    df[history_name] = df[history_name].apply(lambda x: [encode[i] for i in x])
    return decode, len(key)

def Process(dataset, config=None):
    """
        Pre-process dataset, split training-validate-test set into ~/processed_data path

        :param dataset: str: The utilized dataset.
            Note that please download it into ~/dataset path

        :param config: The custom config files.
    """


    with open(os.path.join("recommendation","properties", "dataset.yaml"), "r") as file:
        process_config = yaml.safe_load(file)


    with open(os.path.join("recommendation","properties", "dataset", "{}.yaml".format(dataset)), "r") as file:
        dataset_config = yaml.safe_load(file)

    process_config.update(dataset_config)

    if config:
        process_config.update(config)

    print("process config:")
    print(process_config)


    if os.path.exists(os.path.join("recommendation", "processed_dataset", str(dataset))) and process_config['reprocess'] == False:
        return "do not need to process"

    print("start to process data...")
    inter_file = os.path.join("recommendation","dataset", dataset, dataset+".inter")
    provider_file = os.path.join("recommendation","dataset", dataset, dataset + ".item")


    frames = pd.read_csv(inter_file,delimiter='\t',dtype={process_config["item_id"]:str,process_config["user_id"]:str, process_config['timestamp']:float},
                         usecols=[process_config["user_id"],process_config["item_id"],process_config["label_id"], process_config['timestamp']])
    frames = frames.dropna()
    #provider_frames = pd.read_csv(provider_file, delimiter='\t')
    #print(provider_frames.columns)
    #exit(0)
    provider_frames = pd.read_csv(provider_file,delimiter='\t',
                                  usecols=[process_config["item_id"],process_config["group_id"], process_config["text_id"]],
                                  dtype={process_config["item_id"]:str,process_config["group_id"]:str, process_config["text_id"]:str})

    provider_frames = provider_frames.dropna()


    print(len(provider_frames))

    frames = copy.copy(frames.sample(frac=process_config['sample_size'], random_state=42, axis=0, replace=False).reset_index(drop=True))


    #print("start to pre-process data...")
    uid_field,iid_field,label_field, time_field, text_field = \
        [process_config["user_id"],process_config["item_id"],process_config["label_id"], process_config['timestamp'], process_config["text_id"]]
    id2text = dict(zip(provider_frames[iid_field], provider_frames[text_field]))


    provider_field = process_config["group_id"]

    frames.drop_duplicates(subset=[iid_field,uid_field],keep='first',inplace=True)
    #print(uid_field,iid_field,label_field,time_field)
    item_num, user_num = len(frames[iid_field].unique()),len(frames[uid_field].unique())
    group_num = len(provider_frames[provider_field].unique())
    print("origin:----------item number: %d user number:%d group_num:%d total interactions:%d"%(item_num,user_num,group_num, len(frames)))



    print("start to merge data.....")

    print("processing item val")
    frames[label_field] = frames[label_field].apply(lambda x: 1 if x>=process_config['label_threshold'] else 0)

    frames.rename(columns={label_field:"label:float"},inplace=True)
    label_field = "label:float"
    frames = frames.merge(provider_frames,on=iid_field,how='inner')
    #print(len(frames))
    #exit(0)

    for i in range(3):
        itemLen = frames.groupby(iid_field).size()  # groupby itemID and get size of each item
        remain_items =  (itemLen[itemLen >= process_config['item_val']].index).values
        frames = frames[frames[iid_field].isin(remain_items)].reset_index(drop=True)
        #print(len(frames))
        #print(itemLen)
        userLen = frames.groupby(uid_field).size()
        remain_users =  (userLen[userLen >= process_config['user_val']].index).values
        frames = frames[frames[uid_field].isin(remain_users)].reset_index(drop=True)
        #print(len(frames))
        pidLen = frames.groupby(provider_field)[iid_field].unique().apply(lambda x: len(x))
        remain_providers =  (pidLen[pidLen >= process_config['group_val']].index).values
        frames = frames[frames[provider_field].isin(remain_providers)].reset_index(drop=True)

        #print(len(frames))
        #exit(0)

    frames = frames.sort_values(by=[uid_field, time_field]).reset_index(drop=True)

    def get_previous_items(sub_df):
        clicked_items = []
        previous_items = []

        # go through
        for _, row in sub_df.iterrows():
            #
            previous_items.append(clicked_items[-process_config['history_length']:])  # last k clicked items
            #
            if row[label_field] == 1:
                clicked_items.append(row[iid_field])


        sub_df['history_behaviors'] = previous_items
        return sub_df

    frames = frames.groupby(uid_field, group_keys=False).apply(get_previous_items)
    #print(frames)

    #
    frames = frames[
        frames['history_behaviors'].apply(lambda x: len(x) == process_config['history_length'])].reset_index(drop=True)
    history_field = 'history_behaviors'
    process_config['history_field'] = 'history_behaviors'

    user_num, group_num = len(frames[uid_field].unique()), len(frames[provider_field].unique())





    min_item_size = 99999
    max_item_size = -1

    #build_map(frames,iid_field)
    iid_decode, item_num = build_item_map(frames, iid_field, history_field)
    _ = build_map(frames,uid_field)


    pid2iid = {}
    iid2pid = {}
    for i, (iid, pid) in enumerate(zip(frames[iid_field].values, frames[provider_field].values)):
        if pid not in pid2iid:
            pid2iid[pid] = []
        pid2iid[pid].append(iid)
        iid2pid[iid] = pid
    for key in pid2iid.keys():
        pid2iid[key] = list(set(pid2iid[key]))


    ###Since some group contains few items, we need to aggregate the some groups into one group
    ##SO WE NEED TO remap the provider id
    #index = 1
    remap_dict = {}
    for pid in frames[provider_field].unique():
        items = pid2iid[pid]
        if len(items) < process_config['group_aggregation_threshold']:
            for i in items:
                remap_dict[i] = -1
    iid2pid.update(remap_dict)
    #print(iid2pid)

    frames['remap_pid'] = frames[iid_field].map(iid2pid)
    provider_field = 'remap_pid'
    build_map(frames, provider_field)

    pid2iid = {}
    iid2pid = {}
    for i, (iid, pid) in enumerate(zip(frames[iid_field].values, frames[provider_field].values)):
        iid = str(iid)
        pid = str(pid)
        if pid not in pid2iid:
            pid2iid[pid] = []
        pid2iid[pid].append(iid)
        iid2pid[iid] = pid
    for key in pid2iid.keys():
        pid2iid[key] = str(list(set(pid2iid[key])))

    # item_num, user_num, group_num = len(frames[iid_field].unique()), len(frames[uid_field].unique()), len(
    #     frames[provider_field].unique())
    # print("after process.... item number: %d user number:%d group number:%d total interactions:%d" % (
    # item_num, user_num, group_num, len(frames)))
    group_num = len(frames[provider_field].unique())

    process_config['user_num'] = user_num
    process_config['item_num'] = item_num
    process_config['group_num'] = group_num
    process_config['label_id'] = label_field

    print("item number: %d user number:%d group number:%d total interactions:%d" % (
        item_num, user_num, group_num, len(frames)))



    print("start to sample and split data....")
    frames = frames.sample(frac=1, random_state=42).reset_index(drop=True)
    #frames = frames.reset_index(drop=True)


    #print(frames.head())
    train_items = frames[iid_field].unique()
    train_users = frames[uid_field].unique()
    print("checking.....")


    if np.sum(train_users>=len(train_users)) > 0:
        print("ERROR in user id:",np.sum(train_users>=len(train_users)))
        exit(0)

    print("min user interactions:",np.min(frames.groupby(uid_field).size()))
    print("min item interactions:",np.min(frames.groupby(iid_field).size()))



    final_frame = frames[[uid_field, iid_field, history_field, label_field, provider_field, time_field]]
    final_frame.rename(columns={
        'remap_pid': process_config['group_id']
    }, inplace=True)


    final_frame.sort_values(by=time_field, inplace=True)

    train_size = int(len(final_frame) * (1 - process_config['valid_ratio'] - process_config['test_ratio']))
    val_size = int(len(final_frame) * process_config['valid_ratio'])

    train = final_frame.iloc[:train_size]

    val = final_frame.iloc[train_size:train_size + val_size]
    test = final_frame.iloc[train_size + val_size:]

    print(test.head())
    # train, temp = train_test_split(final_frame, test_size=(1 - process_config['valid_ratio'] - process_config['test_ratio']), random_state=42)
    # val, test = train_test_split(temp, test_size=(process_config['test_ratio'] / (process_config['valid_ratio'] + process_config['test_ratio'])), random_state=42)

    #split train-val-test accoring to the time


    # This is for extracting the positive items for every user to make it easy to conduct negative samples for each users
    user2pos_itemset = {}
    for i in range(len(final_frame)):
        uid = final_frame[uid_field][i]
        iid = final_frame[iid_field][i]
        label = final_frame[label_field][i]
        if uid not in user2pos_itemset.keys():
            user2pos_itemset[uid] = []
        if label == 1:
            user2pos_itemset[uid].append(iid)


    dir = os.path.join("recommendation","processed_dataset",dataset)
    if not os.path.exists(dir):
        os.makedirs(dir)



    #columns = [process_config['user_id'], process_config['item_id'], process_config['label_id'], process_config['group_id']]


    test.to_csv(os.path.join(dir,dataset+".test.CTR"), index=False,sep='\t')
    val.to_csv(os.path.join(dir,dataset+".valid.CTR"), index=False,sep='\t')
    train.to_csv(os.path.join(dir,dataset+".train"), index=False,sep='\t')

    #iid2pid
    def construct_ranking_data(df):

        sort_df = df.sort_values(by=[uid_field, time_field]).reset_index(drop=True)
        #construct_dict = {"user_id":[], "items":[], "pos_length":[] }

        construct_dict = {}
        user_behaviors = {}

        for i in trange(len(sort_df)):

            user, item, label = sort_df[uid_field].values[i], sort_df[iid_field].values[i], sort_df[label_field].values[i]
            history = sort_df[history_field].values[i]
            ##find the behavior sequences in the training dataset
            if label == 1:
                if user not in construct_dict.keys():
                    construct_dict[user] = [item]
                    # filtered_df = train[(train[uid_field] == user) & (train[label_field] == 1)]
                    # print(filtered_df.head())
                    # filtered_df['time_diff'] = abs(filtered_df[time_field] - time)  # find the most close behavior
                    # closest_row = filtered_df.loc[filtered_df['time_diff'].idxmin()]
                    # behavior = closest_row["history_behaviors"][-(process_config['history_length']-1):]
                    # behavior.append(closest_row[iid_field])
                    user_behaviors[user] = history

                else:
                    construct_dict[user].append(item)

        df_dict = {"user_id":[], "history_behaviors":[], "items":[], "pos_length":[] }
        max_len = 0
        for u in tqdm(construct_dict.keys()):
            pos_length = len(construct_dict[u])
            itemset = construct_dict[u]
            if pos_length > max_len:
                max_len = pos_length

            for r in range(process_config['sample_num']-pos_length):
                neg_id = random.randint(0, item_num-1)
                while neg_id in user2pos_itemset[u]:
                    neg_id = random.randint(0, item_num - 1)
                itemset.append(neg_id)

            df_dict["user_id"].append(u)
            df_dict["history_behaviors"].append(user_behaviors[u])
            df_dict["items"].append(itemset)
            df_dict["pos_length"].append(pos_length)

        df_reconstruct = pd.DataFrame(df_dict)
        print(f"max_len:{max_len}")
        return df_reconstruct

    print("start to construct ranking and retrieval test dataset...")
    val_ranking = construct_ranking_data(val)
    test_ranking = construct_ranking_data(test)


    val_ranking.to_csv(os.path.join(dir,dataset+".valid.ranking"), index=False,sep='\t')
    test_ranking.to_csv(os.path.join(dir,dataset+".test.ranking"), index=False,sep='\t')

    id2text_update = {}
    for i in range(item_num):
        ori_i = iid_decode[i]
        text = id2text[ori_i]
        id2text_update[i] = text

    with open(os.path.join(dir, "process_config.yaml"), "w") as file:
        yaml.dump(process_config, file)

    #print(iid2pid)
    with open(os.path.join(dir, "iid2pid.json"), "w") as file:
        json.dump(iid2pid, file)

    with open(os.path.join(dir, "iid2text.json"), "w") as file:
        json.dump(id2text_update, file)


    return f"process complete! The file and config are stored in {dir}"





from sklearn import preprocessing
import os
import pathlib


def create_train_split(df, query_col, id_col, group_col, sort_col, pos_th, data_path, k_fold, ratio_split):
    """
    Creates k-fold train-test splits and saves them as text files.

    :param df: Pandas DataFrame containing the dataset.
    :param query_col: Column representing query IDs (grouping criterion).
    :param id_col: Column containing unique sample IDs.
    :param group_col: Column used for stratified sampling (e.g., demographic group).
    :param sort_col: Column used for positive sample thresholding.
    :param pos_th: Threshold to determine positive samples.
    :param data_path: Directory to save the train-test split files.
    :param k_fold: Number of folds for cross-validation.
    :param ratio_split: Fraction of data used for training in each fold.
    """
    os.makedirs(data_path, exist_ok=True)

    for i in range(k_fold):
        train_ids, test_ids = select_balanced_df(df, query_col, id_col, group_col, sort_col, pos_th, ratio_split)

        # Save train-test splits to files
        with open(os.path.join(data_path, f'test_samples_{i}.txt'), 'w') as f:
            f.write("\n".join(map(str, test_ids)))

        with open(os.path.join(data_path, f'train_samples_{i}.txt'), 'w') as f:
            f.write("\n".join(map(str, train_ids)))


def select_balanced_df(df, query_col, id_col, group_col, sort_col, pos_th, return_ratio):
    """
    Performs stratified sampling to create balanced train-test splits.

    :param df: Pandas DataFrame containing the dataset.
    :param query_col: Column representing query IDs.
    :param id_col: Column containing unique sample IDs.
    :param group_col: Column used for stratification (e.g., demographic group).
    :param sort_col: Column used to separate positive and negative samples.
    :param pos_th: Threshold to classify positive samples.
    :param return_ratio: Fraction of data to use in the training set.
    :return: Tuple (train_samples, test_samples) containing train and test sample IDs.
    """
    train_samples, test_samples, drop_qid = [], [], []

    for qid, df_query in df.groupby(query_col):
        pos_samples = df_query[df_query[sort_col] >= pos_th]  # Positive samples
        neg_samples = df_query[df_query[sort_col] < pos_th]  # Negative samples

        # Identify groups with insufficient samples
        group_sizes = pos_samples[group_col].value_counts()
        insufficient_groups = group_sizes[group_sizes < 2].index.tolist()

        if insufficient_groups:
            drop_qid.append(qid)
            continue  # Skip this query group if it has insufficient data

        # Perform stratified sampling
        stratified_pos_sample = pos_samples.groupby(group_col, group_keys=False).apply(
            lambda x: x.sample(frac=return_ratio, random_state=42))
        stratified_neg_sample = neg_samples.groupby(group_col, group_keys=False).apply(
            lambda x: x.sample(frac=return_ratio, random_state=42))

        train_samples.extend(stratified_pos_sample[id_col].values)
        train_samples.extend(stratified_neg_sample[id_col].values)

        test_samples.extend(df_query.loc[~df_query[id_col].isin(train_samples), id_col].values)

    # Remove queries with insufficient samples for stratified sampling from the train and test sets
    train_samples = [uid for uid in train_samples if uid not in drop_qid]
    test_samples = [uid for uid in test_samples if uid not in drop_qid]

    return train_samples, test_samples



def check_nan(df, cols_train):
    for col in cols_train:
        if df[col].isnull().values.any():
            return True
    return False


def norm_features(features_cols, data):
    # normalize features
    for f in features_cols:
        if min(data[f]) < 0:
            min_value = min(data[f])
            data[f] = data[f].apply(lambda x: x - min_value)

    min_max_scaler = preprocessing.MinMaxScaler()
    data[features_cols] = min_max_scaler.fit_transform(data[features_cols].values)
    return data


def writeToTXT(file_name_with_path, _df):
    # try:
    #     _df.to_csv(file_name_with_path, header=False, index=False, sep=' ')
    # except FileNotFoundError:
    directory = os.path.dirname(file_name_with_path)
    pathlib.Path(directory).mkdir(parents=True, exist_ok=True)
    print("Make folder ", directory)
    _df.to_csv(file_name_with_path, header=False, index=False, sep=' ')


import torch
import torch.nn as nn
import numpy as np
from .abstract_model import AbstractBaseModel


r"""
GRU4Rec
################################################

Reference:
    Yong Kiam Tan et al. "Improved Recurrent Neural Networks for Session-based Recommendations." in DLRS 2016.

"""

class GRU4Rec(AbstractBaseModel):
    def __init__(self, config):
        super().__init__(config)
        # self.user_embedding = nn.Embedding(num_users, embedding_dim)
        # self.item_embedding = nn.Embedding(num_items, embedding_dim)
        self.type = ["sequential"]
        self.IR_type = ["retrieval", "ranking"]

        self.emb_dropout = nn.Dropout(self.config['dropout_prob'])
        self.gru_layers = nn.GRU(
            input_size=self.config['embedding_size'],
            hidden_size=self.config['hidden_size'],
            num_layers=self.config['num_layers'],
            bias=False,
            batch_first=True,
        )
        self.dense = nn.Linear(self.config['hidden_size'], self.config['embedding_size'])

        self.sigmoid = nn.Sigmoid()
        self.loss = nn.CrossEntropyLoss(reduction='none')
        self.apply(self._init_weights)



    def compute_loss(self, interaction):

        #user = interaction['user_ids']
        #history_ids = interaction['history_ids']
        pos_items = interaction['item_ids']
        user_emb = self.get_user_embedding(interaction)
        test_item_emb = self.item_embedding.weight
        logits = torch.matmul(user_emb, test_item_emb.transpose(0, 1))
        loss = self.loss(logits, pos_items)
        return loss

    def get_user_embedding(self, user_dict):
        users = user_dict['history_ids']
        item_seq_emb = self.item_embedding(users)
        item_seq_emb_dropout = self.emb_dropout(item_seq_emb)
        gru_output, _ = self.gru_layers(item_seq_emb_dropout)
        gru_output = self.dense(gru_output)
        seq_output = gru_output[:,-1,:]
        #seq_output = self.gather_indexes(gru_output, self.config['history_length'] - 1)
        return seq_output



    def forward(self, user_dict, item_id):
        ###here user_id denotes the historical ids
        user_embeds = self.get_user_embedding(user_dict)
        item_embeds = self.item_embedding(item_id)
        dot_product = (user_embeds * item_embeds).sum(1)
        return self.sigmoid(dot_product)





from collections import *
import torch
import math
import torch.nn.init as init
from torch import nn
from torch.nn import functional as F
from torch.nn import LayerNorm
import numpy as np

from .base import BasePostProcessModel

MAXDOC=50

'''
@inproceedings{yan2021diversification,
  title={Diversification-aware learning to rank using distributed representation},
  author={Yan, Le and Qin, Zhen and Pasumarthi, Rama Kumar and Wang, Xuanhui and Bendersky, Michael},
  booktitle={Proceedings of the Web Conference 2021},
  pages={127--136},
  year={2021}
}
'''


class MHSA(nn.Module):
    """
    Multi-Head Self-Attention (MHSA) module.

    :param H: The number of attention heads. Default is 2.
    :param input_dim: The input feature dimension. Default is 100.
    :param output_dim: The output feature dimension. Default is 256.
    """
    def __init__(self, H = 2, input_dim = 100, output_dim = 256):
        super(MHSA, self).__init__()
        self.head_num = H
        self.input_dim = input_dim
        self.output_dim = output_dim
        self.norm = LayerNorm(self.output_dim, eps=1e-5)
        self.d_k = math.sqrt(output_dim)
        self.W_Q = nn.ModuleList()
        self.W_K = nn.ModuleList()
        self.W_V = nn.ModuleList()
        self.W_out = nn.Linear(self.output_dim*self.head_num, self.output_dim)
        for i in range(self.head_num):
            self.W_Q.append(nn.Linear(self.input_dim, self.output_dim, False))
            self.W_K.append(nn.Linear(self.input_dim, self.output_dim, False))
            self.W_V.append(nn.Linear(self.input_dim, self.output_dim, False))
        for hid in range(self.head_num):
            init.xavier_normal_(self.W_Q[hid].weight)
            init.xavier_normal_(self.W_K[hid].weight)
            init.xavier_normal_(self.W_V[hid].weight)
        init.xavier_normal_(self.W_out.weight)
    
    def attention(self, Q, K, V):
        """
        Computes scaled dot-product attention.

        :param Q: Query matrix.
        :param K: Key matrix.
        :param V: Value matrix.
        :return: Attention-weighted output.
        """
        scores = torch.bmm(Q, K.transpose(-2,-1)) / self.d_k
        attn = F.softmax(scores, dim = -1)
        return torch.bmm(attn, V)
    
    def forward(self, D):
        """
        Forward pass for the MHSA module.

        :param D: Input tensor of shape (batch_size, sequence_length, input_dim).
        :return: Output tensor after multi-head self-attention and normalization.
        """
        X_list=[]
        for hid in range(self.head_num):
            Q = self.W_Q[hid](D)
            K = self.W_K[hid](D)
            V = self.W_V[hid](D)
            X = self.attention(Q, K, V)
            X_list.append(X)
        res = torch.concat(X_list, dim=-1).float()
        res = self.W_out(res)
        if self.input_dim == self.output_dim:
            res = self.norm(D + res)
        else:
            res = self.norm(res)
        return res


class DALETOR(BasePostProcessModel):
    """
    DALETOR model implementation

    :param dropout: Dropout rate for regularization. Default is 0.1.
    """
    def __init__(self, dropout = 0.1):
        super().__init__(dropout)

        self.activation = "relu"
        self.batch_norm = True
        self.normalization = False
        self.layer_num = 2
        self.feat_dims = [100, 256]
        
        self.DIN1 = MHSA(2, 100, 256)
        self.DIN2 = MHSA(2, 256, 256)
        
        self.fc1 = nn.Linear(556, 256)
        self.fc2 = nn.Linear(256, 128)
        self.fc3 = nn.Linear(128, 64)
        self.fc4 = nn.Linear(64, 1)
        self.batch_norm1 = nn.BatchNorm1d(50)
        self.batch_norm2 = nn.BatchNorm1d(50)
        self.batch_norm3 = nn.BatchNorm1d(50)


        self.nfc1 = nn.Linear(18,18)
        self.nfc2 = nn.Linear(18,8)
        self.nfc3 = nn.Linear(8,1)

        init.xavier_normal_(self.fc1.weight)
        init.xavier_normal_(self.fc2.weight)
        init.xavier_normal_(self.fc3.weight)
        init.xavier_normal_(self.fc4.weight)

        init.xavier_normal_(self.nfc1.weight)
        init.xavier_normal_(self.nfc2.weight)
        init.xavier_normal_(self.nfc3.weight)


    def fit(self, x, rel_feat, train_flag = False):
        """
        Training stage

        :param x: Input tensor of shape (batch_size, sequence_length, feature_dim).
        :param rel_feat: Relevance feature tensor of shape (batch_size, num_docs, 18).
        :param train_flag: Boolean flag indicating whether the model is in training mode.
        :return: Ranking scores for documents.
        """
        bs = x.shape[0]
        seq_len = x.shape[1]
        df = x.shape[2]
        query_x = x[:,0,:].unsqueeze(1)
        doc_x = x[:,1:,:]
        C = query_x * doc_x
        a = self.DIN1(C)
        a = self.DIN2(a)

        query_x = query_x.repeat(1,50,1)
        feat = torch.cat([query_x, doc_x, C, a], dim=2).float()

        s = F.relu(self.batch_norm1(self.fc1(feat)))
        s = F.relu(self.batch_norm2(self.fc2(s)))
        s = F.relu(self.batch_norm3(self.fc3(s)))
        s = self.fc4(s)
        s = s.squeeze()

        rel_feat = rel_feat.reshape(rel_feat.shape[0] * rel_feat.shape[1], 18)
        sr = F.relu(self.nfc1(rel_feat))
        sr = F.relu(self.nfc2(sr))
        sr = self.nfc3(sr)
        sr = sr.squeeze()
        sr = sr.reshape(bs, MAXDOC)

        score = s + sr
        
        if train_flag:
            return score
        else:
            return score



import numpy as np
import os
import yaml
from scipy.sparse import save_npz, load_npz
from .rerank_model import CPFair, FairRec, FairRecPlus, k_neighbor, min_regularizer, PMMF, Welf, TaxRank, FairSync, RAIF
from .metric import dcg, MMF, Gini, Entropy, MinMaxRatio
from datetime import datetime
import json


class RecReRanker(object):
    def __init__(self, train_config):
        """Initialize In-processing and base models.

        :param train_config: Your custom config files.
        """


        self.dataset = train_config['dataset']
        #self.stage = stage
        self.train_config = train_config


    def load_configs(self, dir):
        """
           Loads and merges configuration files for the model, dataset, and evaluation.

           This function loads multiple YAML configuration files, including the process configuration,
           dataset-specific settings, model configurations, and evaluation parameters. All configurations
           are merged, with the highest priority given to the class's own `config` attribute.

           :param dir: The directory where the main process configuration file is located.
           :return: A dictionary containing the merged configuration from all files.
        """
        print("start to load config...")
        with open(os.path.join(dir, "process_config.yaml"), 'r') as f:
            config = yaml.safe_load(f)
        # print(train_data_df.head())

        print("start to load model...")
        with open(os.path.join("recommendation", "properties", "models.yaml"), 'r') as f:
            model_config = yaml.safe_load(f)

        model_path = os.path.join("recommendation", "properties", "models", self.train_config['model'] + ".yaml")
        # if not os.path.exists(model_path):
        #     raise NotImplementedError("we do not support such model type!")
        with open(model_path, 'r') as f:
            model_config.update(yaml.safe_load(f))
        config.update(model_config)

        with open(os.path.join("recommendation", "properties", "evaluation.yaml"), 'r') as f:
            config.update(yaml.safe_load(f))

        config.update(self.train_config)  ###train_config has highest rights
        print("your loading config is:")
        print(config)

        return config

    def rerank(self):
        """
            Training post-processing main workflow.
        """

        dir = os.path.join("recommendation", "processed_dataset", self.dataset)
        config = self.load_configs(dir)

        ranking_score_path = os.path.join("recommendation", "log", config['ranking_store_path'])
        if not os.path.exists(ranking_score_path):
            raise ValueError(f"do not exist the path {ranking_score_path}, please check the path or run the ranking phase to generate scores for re-ranking !")
        print("loading ranking scores....")
        file = os.path.join(ranking_score_path, "ranking_scores.npz")
        ranking_scores = load_npz(file).toarray() #[user_num, item_num]
        ###we need to remove the group do not appear after the ranking phase, for evaluate full group, please evaluate in retrieval stage



        if config['model'] == "CPFair":
            Reranker = CPFair(config)
        elif config['model'] == "FairRec":
            Reranker = FairRec(config)
        elif config['model'] == "FairRecPlus":
            Reranker = FairRecPlus(config)
        elif config['model'] == 'k_neighbor':
            Reranker = k_neighbor(config)
        elif config['model'] == 'min_regularizer':
            Reranker = min_regularizer(config)
        elif config['model'] == 'PMMF':
            Reranker = PMMF(config)
        elif config['model'] == 'Welf':
            Reranker = Welf(config)
        elif config['model'] == 'TaxRank':
            Reranker = TaxRank(config)
        elif config['model'] == 'FairSync':
            Reranker = FairSync(config)
        elif config['model'] == 'RAIF':
            Reranker = RAIF(config)
        else:
            raise NotImplementedError(f"We do not support the model type {self.train_config['model']}")

        #item_scores = np.sum(ranking_scores, axis=0, keepdims=False)

        metrics = ["ndcg", "u_loss"]
        rerank_result = {}
        exposure_result = {}
        for k in config['topk']:
            rerank_result.update({f"{m}@{k}":0 for m in metrics})

            rerank_list = Reranker.rerank(ranking_scores, k)
            exposure_list = np.zeros(config['group_num'])
            for u in range(len(rerank_list)):
                sorted_result_score = np.sort(ranking_scores[u])[::-1]
                true_dcg = dcg(sorted_result_score, k)
                rerank_items = rerank_list[u]

                for i in rerank_items:
                    if i not in Reranker.iid2pid.keys():
                        gid = 0
                    else:
                        gid = Reranker.iid2pid[i]
                    if config['fairness_type'] == "Exposure":
                        exposure_list[gid] += 1
                    else:
                        exposure_list[gid] += np.round(ranking_scores[u][i], config['decimals'])
                reranked_score = ranking_scores[u][rerank_items]
                pre_dcg = dcg(np.sort(reranked_score)[::-1], k)
                rerank_result[f"ndcg@{k}"] += pre_dcg/true_dcg
                rerank_result[f"u_loss@{k}"] += (np.sum(sorted_result_score[:k]) - np.sum(reranked_score[:k]))/k

            rerank_result[f"ndcg@{k}"] /= len(rerank_list)
            rerank_result[f"u_loss@{k}"] /= len(rerank_list)
            for fairness_metric in self.train_config['fairness_metrics']:
                if fairness_metric == 'MinMaxRatio':
                    rerank_result[f"MinMaxRatio@{k}"] = MinMaxRatio(exposure_list)
                elif fairness_metric == 'MMF':
                    rerank_result[f"MMF@{k}"] = MMF(exposure_list)
                elif fairness_metric == 'Entropy':
                    rerank_result[f"Entropy@{k}"] = Entropy(exposure_list)
                elif fairness_metric == 'GINI':
                    rerank_result[f"GINI@{k}"] = Gini(exposure_list)

            #rerank_result[f"mmf@{k}"] = MMF(exposure_list)
            #rerank_result[f"gini@{k}"] = Gini(exposure_list)
            #print(exposure_list)
            exposure_result[f"top@{k}"] = str(list(exposure_list))


        for k in rerank_result.keys():
            rerank_result[k] = np.round(rerank_result[k], config['decimals'])


        today = datetime.today()
        today_str = f"{today.year}-{today.month}-{today.day}"
        log_dir = os.path.join("recommendation", "log", f"{today_str}_{config['log_name']}")

        if not os.path.exists(log_dir):
            os.makedirs(log_dir)
        with open(os.path.join(log_dir, 'test_result.json'), 'w') as file:
            json.dump(rerank_result, file)
        with open(os.path.join(log_dir, 'exposure_result.json'), 'w') as file:
            json.dump(exposure_result, file)
            #file.write(str(exposure_list))
        #print("exposure list:")
        #print(exposure_list)
        print(rerank_result)

        with open(os.path.join(log_dir, "config.yaml"), 'w') as f:
            yaml.dump(config, f)

        print(f"result and config dump in {log_dir}")









import numpy as np
from .Abstract_Ranker import Abstract_Reweigher

r"""
APR
################################################


@INPROCEEDINGS{APR,
  author={Hu, Zhihao and Xu, Yiran and Tian, Xinmei},
  booktitle={2023 International Joint Conference on Neural Networks (IJCNN)}, 
  title={Adaptive Priority Reweighing for Generalizing Fairness Improvement}, 
  year={2023},
  volume={},
  number={},
  pages={01-08},
  keywords={Training;Adaptation models;Neural networks;Decision making;Machine learning;Predictive models;Benchmark testing},
  doi={10.1109/IJCNN54540.2023.10191757}
}



"""

class APR(Abstract_Reweigher):
    def __init__(self, config, group_weight):
        super().__init__(config)
        self.group_weight = group_weight

    def reset_parameters(self):
        self.sigma = np.zeros(self.config['group_num'])
        #self.subgroup_weights = np.ones(self.config['group_num']) * 0.1

    def reweight(self, input_dict):
        items = input_dict['items']
        losses = input_dict['loss'] #[B]

        adj_matrix = self.M[items]
        group_num = np.sum(adj_matrix, axis=0)
        loss_groups = np.sum(losses[:, np.newaxis] * adj_matrix, axis=0, keepdims=False)/(group_num+1e-1)
        #loss_groups = loss_groups / np.sum(loss_groups)

        #print(loss_groups)
        #exit(0)

        self.sigma = (1-self.config['beta']) + self.config['beta'] * (loss_groups-0.0)  #here since we utilize the loss, therefore d=0
        #print(self.sigma)
        B_t = np.sum(adj_matrix, axis=0, keepdims=False)
        #self.subgroup_weights = self.subgroup_weights * 1/((B_t+self.config['alpha'])/(np.sum(B_t)+self.config['alpha']))
        exp_sigma = 1/np.exp(-self.config['eta']*self.sigma)
        #weights = self.subgroup_weights * exp_sigma
        weights = exp_sigma
        batch_weight = np.matmul(adj_matrix, weights * self.group_weight)
        batch_weight = batch_weight / np.sum(batch_weight)
        #print(batch_weight)
        #exit(0)

        return batch_weight




import numpy as np
import pandas as pd
import os
def compute_dynamic_values_to_code(data, sensitive_attributes):
    value_to_code = {}
    current_code = 1

    # For each sensitive attribute
    for s_attribute in sorted(sensitive_attributes):
        unique_groups = sorted(set(data[s_attribute].unique()))

        # Assign code to independent groups
        for group in unique_groups:
            if group not in value_to_code:
                value_to_code[group] = current_code
                current_code += 1

    return value_to_code




def process_data_input(data, configs, dataset):
    features_col = dataset.feature_cols + [dataset.score_col]

    if "unprivileged_groups" in configs:
        sensitive_attributes = set(configs["unprivileged_groups"]) | set(configs["privileged_groups"])
        value_to_code = compute_dynamic_values_to_code(data, sensitive_attributes)

        unprivileged_groups, privileged_groups, group_weights = {}, {}, {}
        # Encode sensitive attributes
        for s_attribute in sensitive_attributes:
            coded_attribute = s_attribute + '_coded'
            features_col.append(coded_attribute)

            data[coded_attribute] = data[s_attribute].map(value_to_code)

            unprivileged_groups[s_attribute] = [value_to_code[group] for group in
                                                configs["unprivileged_groups"].get(s_attribute, [])]
            privileged_groups[s_attribute] = [value_to_code[group] for group in
                                              configs["privileged_groups"].get(s_attribute, [])]

            # Update group weights
            for group in value_to_code:
                if group in data[s_attribute].unique() and value_to_code[group] not in group_weights:
                    group_weights[value_to_code[group]] = configs['group_weights'].get(group, 0)

        # Store sensitive groups
        sensitive_groups = {"unprivileged_groups": unprivileged_groups, "privileged_groups": privileged_groups}
    else:
        sensitive_attributes = dataset.sensitive_cols
        value_to_code = compute_dynamic_values_to_code(data, sensitive_attributes)
        for s_attribute in sensitive_attributes:
            coded_attribute = s_attribute + '_coded'
            features_col.append(coded_attribute)

            data[coded_attribute] = data[s_attribute].map(value_to_code)

        group_weights = None
        sensitive_groups = None

    # Identify indices of sensitive columns
    sensitive_column_indices = [
        list(data[features_col].columns).index(s_attribute + "_coded") for s_attribute in sensitive_attributes
    ]

    nonsensitive_column_indices = [index for index in range(0, len(features_col)) if
                                   index not in sensitive_column_indices]
    data_processed = data[features_col].to_numpy()
    return data_processed, group_weights, sensitive_groups, sensitive_column_indices, nonsensitive_column_indices


def process_data_output(data_orig, data_fair, dataset, nonsensitive_column_indices, fair_data_path, file_name=None):
    # Combine relevant columns from configs
    features_col = dataset.feature_cols + [dataset.score_col]

    # Convert data_fair to a numpy array and select relevant columns
    data_fair = np.vstack(data_fair)
    fair_features_col = [f'{col}_fair' for col in features_col]
    selected_fair_features_col = [fair_features_col[i] for i in nonsensitive_column_indices]

    # Create DataFrame for fair data with selected columns
    data_fair = pd.DataFrame(data_fair[:, nonsensitive_column_indices], columns=selected_fair_features_col)

    # Add original data columns to the fair data
    for col in data_orig.columns:
        data_fair[col] = data_orig[col].values
        data_fair[col] = data_orig[col].values

    if file_name != None:
        os.makedirs(fair_data_path, exist_ok=True)
        data_fair.to_csv(os.path.join(fair_data_path, f'fair_{file_name}_data.csv'))

    return data_fair


def save_model_data(model, path):
    with open(os.path.join(path, 'model_parmas.npy'), 'wb') as f:
        np.save(f, model.opt_params)


def load_model_data(path):
    with open(os.path.join(path, 'model_parmas.npy'), 'rb') as f:
       return np.load(f)




import numpy as np
import math
from .Abstract_Reranker import Abstract_Reranker
from tqdm import tqdm,trange

import gzip,pickle
import random,math
from itertools import permutations
import sys,datetime
import networkx as nx

'''
FairRec+ Model for item-side fair re-ranking improved by the FairRec Model

@article{biswas2021toward,
  title={Toward Fair Recommendation in Two-sided Platforms},
  author={Biswas, Arpita and Patro, Gourab K and Ganguly, Niloy and Gummadi, Krishna P and Chakraborty, Abhijnan},
  journal={ACM Transactions on the Web (TWEB)},
  volume={16},
  number={2},
  pages={1--34},
  year={2021},
  publisher={ACM New York, NY}
}
'''


def remove_envy_cycle(B,U,V):
    r=0
    while True:
        r+=1
        #print("In envy cycle removal:",r)
        # create empty graph
        G=nx.DiGraph()
        # add nodes
        G.add_nodes_from(U)
        # edges
        E=[]
        # find edges
        #print("In envy cycle removal: finding edges")
        for u in U:
            for v in U:
                if u!=v:
                    V_u=0
                    V_v=0
                    for p in B[u]:
                        V_u+=V[u,p]
                    for p in B[v]:
                        V_v+=V[u,p]
                    if V_v>V_u:
                        E.append((u,v))
        # add edges to the graph
        G.add_edges_from(E)
        # find cycle and remove
        #print("In envy cycle removal: graph done, finding and removing cycles")
        try:
            cycle=nx.find_cycle(G,orientation="original")
            temp=B[cycle[0][0]][:]
            for pair in cycle:
                B[pair[0]]=B[pair[1]][:]
            B[cycle[-1][0]]=temp[:]
        except:
            break
    # topological sort
    U=list(nx.topological_sort(G))
    return B.copy(),U[:]

# greedy round robin allocation based on a specific ordering of customers
# This is the modified greedy round robin where we remove envy cycles
def greedy_round_robin(m,n,R,l,T,V,U,F):
    # creating empty allocations
    B={}
    for u in U:
        B[u]=[]

    # available number of copies of each producer
    Z={} # total availability
    P=range(n) # set of producers
    for p in P:
        Z[p]=l

    # number of rounds
    r=0
    while True:
        # number of rounds
        r=r+1
        if r > 200:
            break
        # allocating the producers to customers
        #print("GRR round number==============================",r)

        for i in range(m):
            #user
            u=U[i]

            # choosing the p_ which is available and also in feasible set for the user
            possible=[(Z[p]>0)*(p in F[u])*V[u,p] for p in range(n)]
            p_=np.argmax(possible)

            if (Z[p_]>0) and (p_ in F[u]) and len(F[u])>0:
                B[u].append(p_)
                F[u].remove(p_)
                Z[p_]=Z[p_]-1
                T=T-1


            else: #stopping criteria
                #print("now doing envy cycle removal")
                B,U=remove_envy_cycle(B.copy(),U[:],V)
                return B.copy(),F.copy()

            if T==0: #stopping criteria
                #print("now doing envy cycle removal")
                B,U=remove_envy_cycle(B.copy(),U[:],V)
                return B.copy(),F.copy()
        # envy-based manipulations, m, U, V, B.copy()
        #print("GRR done")

        # remove envy cycle
        #print("now doing envy cycle removal")
        B,U=remove_envy_cycle(B.copy(),U[:],V)
        #print(sum([len(B[u]) for u in B]),T,n*l)
    # returning the allocation
    return B.copy(),F.copy()


def FairRecPlus_train(U,P,k,V,alpha, m, n):
    # Allocation set for each customer, initially it is set to empty set
    A={}
    for u in U:
        A[u]=[]

    # feasible set for each customer, initially it is set to P
    F={}
    for u in U:
        F[u]=P[:]
    #print(sum([len(F[u]) for u in U]))

    # l= number of copies of each producer, equal to the exposure guarantee for producers
    l=int(alpha*m*k/(n+0.0))

    # R= number of rounds of allocation to be done in first GRR
    R=int(math.ceil((l*n)/(m+0.0)))


    # T= total number of products to be allocated
    T= l*n

    # first greedy round-robin allocation
    B={}
    [B,F1]=greedy_round_robin(m,n,R,l,T,V,U[:],F.copy())
    F={}
    F=F1.copy()
    #print("GRR done")
    # adding the allocation
    for u in U:
        A[u]=A[u][:]+B[u][:]


    # filling the recommendation set upto size k
    u_less=[]
    for u in A:
        if len(A[u])<k:
            u_less.append(u)
    for u in u_less:
        scores=V[u,:]
        new=scores.argsort()[-(k+k):][::-1]
        for p in new:
            if p not in A[u]:
                A[u].append(p)
            if len(A[u])==k:
                break
    end_time=datetime.datetime.now()
    return A

class FairRecPlus(Abstract_Reranker):
    def __init__(self, config, weights=None):
        super().__init__(config, weights)

    def rerank(self, ranking_score, k):

        user_size = len(ranking_score)
        U=list(range(user_size)) # list of customers
        P=list(range(self.config['item_num'])) # list of producers
        rerank_list =FairRecPlus_train(U,P,k, ranking_score ,alpha=self.config['para'], m=user_size, n=self.config['item_num'])
        return rerank_list

        #rho_reverse = 1/(self.rho*batch_size*self.TopK)





# -*- coding: utf-8 -*-



import copy
import math

import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as fn
from torch.nn.init import normal_



class MLPLayers(nn.Module):
    r"""MLPLayers

    Args:
        - layers(list): a list contains the size of each layer in mlp layers
        - dropout(float): probability of an element to be zeroed. Default: 0
        - activation(str): activation function after each layer in mlp layers. Default: 'relu'.
                           candidates: 'sigmoid', 'tanh', 'relu', 'leekyrelu', 'none'

    Shape:

        - Input: (:math:`N`, \*, :math:`H_{in}`) where \* means any number of additional dimensions
          :math:`H_{in}` must equal to the first value in `layers`
        - Output: (:math:`N`, \*, :math:`H_{out}`) where :math:`H_{out}` equals to the last value in `layers`

    Examples::

        >>> m = MLPLayers([64, 32, 16], 0.2, 'relu')
        >>> input = torch.randn(128, 64)
        >>> output = m(input)
        >>> print(output.size())
        >>> torch.Size([128, 16])
    """

    def __init__(
        self,
        layers,
        dropout=0.0,
        activation="relu",
        bn=False,
        init_method=None,
        last_activation=True,
    ):
        super(MLPLayers, self).__init__()
        self.layers = layers
        self.dropout = dropout
        self.activation = activation
        self.use_bn = bn
        self.init_method = init_method

        mlp_modules = []
        for idx, (input_size, output_size) in enumerate(
            zip(self.layers[:-1], self.layers[1:])
        ):
            mlp_modules.append(nn.Dropout(p=self.dropout))
            mlp_modules.append(nn.Linear(input_size, output_size))
            if self.use_bn:
                mlp_modules.append(nn.BatchNorm1d(num_features=output_size))
            activation_func = activation_layer(self.activation, output_size)
            if activation_func is not None:
                mlp_modules.append(activation_func)
        if self.activation is not None and not last_activation:
            mlp_modules.pop()
        self.mlp_layers = nn.Sequential(*mlp_modules)
        if self.init_method is not None:
            self.apply(self.init_weights)

    def init_weights(self, module):
        # We just initialize the module with normal distribution as the paper said
        if isinstance(module, nn.Linear):
            if self.init_method == "norm":
                normal_(module.weight.data, 0, 0.01)
            if module.bias is not None:
                module.bias.data.fill_(0.0)

    def forward(self, input_feature):
        return self.mlp_layers(input_feature)


def activation_layer(activation_name="relu", emb_dim=None):
    """Construct activation layers

    Args:
        activation_name: str, name of activation function
        emb_dim: int, used for Dice activation

    Return:
        activation: activation layer
    """
    if activation_name is None:
        activation = None
    elif isinstance(activation_name, str):
        if activation_name.lower() == "sigmoid":
            activation = nn.Sigmoid()
        elif activation_name.lower() == "tanh":
            activation = nn.Tanh()
        elif activation_name.lower() == "relu":
            activation = nn.ReLU()
        elif activation_name.lower() == "leakyrelu":
            activation = nn.LeakyReLU()
        elif activation_name.lower() == "dice":
            activation = Dice(emb_dim)
        elif activation_name.lower() == "none":
            activation = None
    elif issubclass(activation_name, nn.Module):
        activation = activation_name()
    else:
        raise NotImplementedError(
            "activation function {} is not implemented".format(activation_name)
        )

    return activation




class FLEmbedding(nn.Module):
    r"""Embedding for float fields.

    Args:
        field_dims: list, the number of float in each float fields
        offsets: list, the dimension offset of each float field
        embed_dim: int, the dimension of output embedding vectors

    Input:
        input_x: tensor, A 3D tensor with shape:``(batch_size,field_size,2)``.

    Return:
        output: tensor,  A 3D tensor with shape: ``(batch_size,field_size,embed_dim)``.
    """

    def __init__(self, field_dims, offsets, embed_dim):
        super(FLEmbedding, self).__init__()
        self.embedding = nn.Embedding(sum(field_dims), embed_dim)
        self.offsets = offsets

    def forward(self, input_x):
        base, index = torch.split(input_x, [1, 1], dim=-1)
        index = index.squeeze(-1).long()
        index = index + index.new_tensor(self.offsets).unsqueeze(0)
        output = base * self.embedding(index)
        return output


class BaseFactorizationMachine(nn.Module):
    r"""Calculate FM result over the embeddings

    Args:
        reduce_sum: bool, whether to sum the result, default is True.

    Input:
        input_x: tensor, A 3D tensor with shape:``(batch_size,field_size,embed_dim)``.

    Output
        output: tensor, A 3D tensor with shape: ``(batch_size,1)`` or ``(batch_size, embed_dim)``.
    """

    def __init__(self, reduce_sum=True):
        super(BaseFactorizationMachine, self).__init__()
        self.reduce_sum = reduce_sum

    def forward(self, input_x):
        square_of_sum = torch.sum(input_x, dim=1) ** 2
        sum_of_square = torch.sum(input_x**2, dim=1)
        output = square_of_sum - sum_of_square
        if self.reduce_sum:
            output = torch.sum(output, dim=1, keepdim=True)
        output = 0.5 * output
        return output


class BiGNNLayer(nn.Module):
    r"""Propagate a layer of Bi-interaction GNN

    .. math::
        output = (L+I)EW_1 + LE \otimes EW_2
    """

    def __init__(self, in_dim, out_dim):
        super(BiGNNLayer, self).__init__()
        self.in_dim = in_dim
        self.out_dim = out_dim
        self.linear = torch.nn.Linear(in_features=in_dim, out_features=out_dim)
        self.interActTransform = torch.nn.Linear(
            in_features=in_dim, out_features=out_dim
        )

    def forward(self, lap_matrix, eye_matrix, features):
        # for GCF ajdMat is a (N+M) by (N+M) mat
        # lap_matrix L = D^-1(A)D^-1 # 拉普拉斯矩阵
        x = torch.sparse.mm(lap_matrix, features)

        inter_part1 = self.linear(features + x)
        inter_feature = torch.mul(x, features)
        inter_part2 = self.interActTransform(inter_feature)

        return inter_part1 + inter_part2


class AttLayer(nn.Module):
    """Calculate the attention signal(weight) according the input tensor.

    Args:
        infeatures (torch.FloatTensor): A 3D input tensor with shape of[batch_size, M, embed_dim].

    Returns:
        torch.FloatTensor: Attention weight of input. shape of [batch_size, M].
    """

    def __init__(self, in_dim, att_dim):
        super(AttLayer, self).__init__()
        self.in_dim = in_dim
        self.att_dim = att_dim
        self.w = torch.nn.Linear(in_features=in_dim, out_features=att_dim, bias=False)
        self.h = nn.Parameter(torch.randn(att_dim), requires_grad=True)

    def forward(self, infeatures):
        att_signal = self.w(infeatures)  # [batch_size, M, att_dim]
        att_signal = fn.relu(att_signal)  # [batch_size, M, att_dim]

        att_signal = torch.mul(att_signal, self.h)  # [batch_size, M, att_dim]
        att_signal = torch.sum(att_signal, dim=2)  # [batch_size, M]
        att_signal = fn.softmax(att_signal, dim=1)  # [batch_size, M]

        return att_signal


class Dice(nn.Module):
    r"""Dice activation function

    .. math::
        f(s)=p(s) \cdot s+(1-p(s)) \cdot \alpha s

    .. math::
        p(s)=\frac{1} {1 + e^{-\frac{s-E[s]} {\sqrt {Var[s] + \epsilon}}}}
    """

    def __init__(self, emb_size):
        super(Dice, self).__init__()

        self.sigmoid = nn.Sigmoid()
        self.alpha = torch.zeros((emb_size,))

    def forward(self, score):
        self.alpha = self.alpha.to(score.device)
        score_p = self.sigmoid(score)

        return self.alpha * (1 - score_p) * score + score_p * score


class SequenceAttLayer(nn.Module):
    """Attention Layer. Get the representation of each user in the batch.

    Args:
        queries (torch.Tensor): candidate ads, [B, H], H means embedding_size * feat_num
        keys (torch.Tensor): user_hist, [B, T, H]
        keys_length (torch.Tensor): mask, [B]

    Returns:
        torch.Tensor: result
    """

    def __init__(
        self,
        mask_mat,
        att_hidden_size=(80, 40),
        activation="sigmoid",
        softmax_stag=False,
        return_seq_weight=True,
    ):
        super(SequenceAttLayer, self).__init__()
        self.att_hidden_size = att_hidden_size
        self.activation = activation
        self.softmax_stag = softmax_stag
        self.return_seq_weight = return_seq_weight
        self.mask_mat = mask_mat
        self.att_mlp_layers = MLPLayers(
            self.att_hidden_size, activation=self.activation, bn=False
        )
        self.dense = nn.Linear(self.att_hidden_size[-1], 1)

    def forward(self, queries, keys, keys_length):
        embedding_size = queries.shape[-1]  # H
        hist_len = keys.shape[1]  # T
        queries = queries.repeat(1, hist_len)

        queries = queries.view(-1, hist_len, embedding_size)

        # MLP Layer
        input_tensor = torch.cat(
            [queries, keys, queries - keys, queries * keys], dim=-1
        )
        output = self.att_mlp_layers(input_tensor)
        output = torch.transpose(self.dense(output), -1, -2)

        # get mask
        output = output.squeeze(1)
        mask = self.mask_mat.repeat(output.size(0), 1)
        mask = mask >= keys_length.unsqueeze(1)

        # mask
        if self.softmax_stag:
            mask_value = -np.inf
        else:
            mask_value = 0.0

        output = output.masked_fill(mask=mask, value=torch.tensor(mask_value))
        output = output.unsqueeze(1)
        output = output / (embedding_size**0.5)

        # get the weight of each user's history list about the target item
        if self.softmax_stag:
            output = fn.softmax(output, dim=2)  # [B, 1, T]

        if not self.return_seq_weight:
            output = torch.matmul(output, keys)  # [B, 1, H]

        return output


class VanillaAttention(nn.Module):
    """
    Vanilla attention layer is implemented by linear layer.

    Args:
        input_tensor (torch.Tensor): the input of the attention layer

    Returns:
        hidden_states (torch.Tensor): the outputs of the attention layer
        weights (torch.Tensor): the attention weights

    """

    def __init__(self, hidden_dim, attn_dim):
        super().__init__()
        self.projection = nn.Sequential(
            nn.Linear(hidden_dim, attn_dim), nn.ReLU(True), nn.Linear(attn_dim, 1)
        )

    def forward(self, input_tensor):
        # (B, Len, num, H) -> (B, Len, num, 1)
        energy = self.projection(input_tensor)
        weights = torch.softmax(energy.squeeze(-1), dim=-1)
        # (B, Len, num, H) * (B, Len, num, 1) -> (B, len, H)
        hidden_states = (input_tensor * weights.unsqueeze(-1)).sum(dim=-2)
        return hidden_states, weights


class MultiHeadAttention(nn.Module):
    """
    Multi-head Self-attention layers, a attention score dropout layer is introduced.

    Args:
        input_tensor (torch.Tensor): the input of the multi-head self-attention layer
        attention_mask (torch.Tensor): the attention mask for input tensor

    Returns:
        hidden_states (torch.Tensor): the output of the multi-head self-attention layer

    """

    def __init__(
        self,
        n_heads,
        hidden_size,
        hidden_dropout_prob,
        attn_dropout_prob,
        layer_norm_eps,
    ):
        super(MultiHeadAttention, self).__init__()
        if hidden_size % n_heads != 0:
            raise ValueError(
                "The hidden size (%d) is not a multiple of the number of attention "
                "heads (%d)" % (hidden_size, n_heads)
            )

        self.num_attention_heads = n_heads
        self.attention_head_size = int(hidden_size / n_heads)
        self.all_head_size = self.num_attention_heads * self.attention_head_size
        self.sqrt_attention_head_size = math.sqrt(self.attention_head_size)

        self.query = nn.Linear(hidden_size, self.all_head_size)
        self.key = nn.Linear(hidden_size, self.all_head_size)
        self.value = nn.Linear(hidden_size, self.all_head_size)

        self.softmax = nn.Softmax(dim=-1)
        self.attn_dropout = nn.Dropout(attn_dropout_prob)

        self.dense = nn.Linear(hidden_size, hidden_size)
        self.LayerNorm = nn.LayerNorm(hidden_size, eps=layer_norm_eps)
        self.out_dropout = nn.Dropout(hidden_dropout_prob)

    def transpose_for_scores(self, x):
        new_x_shape = x.size()[:-1] + (
            self.num_attention_heads,
            self.attention_head_size,
        )
        x = x.view(*new_x_shape)
        return x

    def forward(self, input_tensor, attention_mask):
        mixed_query_layer = self.query(input_tensor)
        mixed_key_layer = self.key(input_tensor)
        mixed_value_layer = self.value(input_tensor)

        query_layer = self.transpose_for_scores(mixed_query_layer).permute(0, 2, 1, 3)
        key_layer = self.transpose_for_scores(mixed_key_layer).permute(0, 2, 3, 1)
        value_layer = self.transpose_for_scores(mixed_value_layer).permute(0, 2, 1, 3)

        # Take the dot product between "query" and "key" to get the raw attention scores.
        attention_scores = torch.matmul(query_layer, key_layer)

        attention_scores = attention_scores / self.sqrt_attention_head_size
        # Apply the attention mask is (precomputed for all layers in BertModel forward() function)
        # [batch_size heads seq_len seq_len] scores
        # [batch_size 1 1 seq_len]
        attention_scores = attention_scores + attention_mask

        # Normalize the attention scores to probabilities.
        attention_probs = self.softmax(attention_scores)
        # This is actually dropping out entire tokens to attend to, which might
        # seem a bit unusual, but is taken from the original Transformer paper.

        attention_probs = self.attn_dropout(attention_probs)
        context_layer = torch.matmul(attention_probs, value_layer)
        context_layer = context_layer.permute(0, 2, 1, 3).contiguous()
        new_context_layer_shape = context_layer.size()[:-2] + (self.all_head_size,)
        context_layer = context_layer.view(*new_context_layer_shape)
        hidden_states = self.dense(context_layer)
        hidden_states = self.out_dropout(hidden_states)
        hidden_states = self.LayerNorm(hidden_states + input_tensor)

        return hidden_states


class FeedForward(nn.Module):
    """
    Point-wise feed-forward layer is implemented by two dense layers.

    Args:
        input_tensor (torch.Tensor): the input of the point-wise feed-forward layer

    Returns:
        hidden_states (torch.Tensor): the output of the point-wise feed-forward layer

    """

    def __init__(
        self, hidden_size, inner_size, hidden_dropout_prob, hidden_act, layer_norm_eps
    ):
        super(FeedForward, self).__init__()
        self.dense_1 = nn.Linear(hidden_size, inner_size)
        self.intermediate_act_fn = self.get_hidden_act(hidden_act)

        self.dense_2 = nn.Linear(inner_size, hidden_size)
        self.LayerNorm = nn.LayerNorm(hidden_size, eps=layer_norm_eps)
        self.dropout = nn.Dropout(hidden_dropout_prob)

    def get_hidden_act(self, act):
        ACT2FN = {
            "gelu": self.gelu,
            "relu": fn.relu,
            "swish": self.swish,
            "tanh": torch.tanh,
            "sigmoid": torch.sigmoid,
        }
        return ACT2FN[act]

    def gelu(self, x):
        """Implementation of the gelu activation function.

        For information: OpenAI GPT's gelu is slightly different (and gives slightly different results)::

            0.5 * x * (1 + torch.tanh(math.sqrt(2 / math.pi) * (x + 0.044715 * torch.pow(x, 3))))

        Also see https://arxiv.org/abs/1606.08415
        """
        return x * 0.5 * (1.0 + torch.erf(x / math.sqrt(2.0)))

    def swish(self, x):
        return x * torch.sigmoid(x)

    def forward(self, input_tensor):
        hidden_states = self.dense_1(input_tensor)
        hidden_states = self.intermediate_act_fn(hidden_states)

        hidden_states = self.dense_2(hidden_states)
        hidden_states = self.dropout(hidden_states)
        hidden_states = self.LayerNorm(hidden_states + input_tensor)

        return hidden_states


class TransformerLayer(nn.Module):
    """
    One transformer layer consists of a multi-head self-attention layer and a point-wise feed-forward layer.

    Args:
        hidden_states (torch.Tensor): the input of the multi-head self-attention sublayer
        attention_mask (torch.Tensor): the attention mask for the multi-head self-attention sublayer

    Returns:
        feedforward_output (torch.Tensor): The output of the point-wise feed-forward sublayer,
                                           is the output of the transformer layer.

    """

    def __init__(
        self,
        n_heads,
        hidden_size,
        intermediate_size,
        hidden_dropout_prob,
        attn_dropout_prob,
        hidden_act,
        layer_norm_eps,
    ):
        super(TransformerLayer, self).__init__()
        self.multi_head_attention = MultiHeadAttention(
            n_heads, hidden_size, hidden_dropout_prob, attn_dropout_prob, layer_norm_eps
        )
        self.feed_forward = FeedForward(
            hidden_size,
            intermediate_size,
            hidden_dropout_prob,
            hidden_act,
            layer_norm_eps,
        )

    def forward(self, hidden_states, attention_mask):
        attention_output = self.multi_head_attention(hidden_states, attention_mask)
        feedforward_output = self.feed_forward(attention_output)
        return feedforward_output


class TransformerEncoder(nn.Module):
    r"""One TransformerEncoder consists of several TransformerLayers.

    Args:
        n_layers(num): num of transformer layers in transformer encoder. Default: 2
        n_heads(num): num of attention heads for multi-head attention layer. Default: 2
        hidden_size(num): the input and output hidden size. Default: 64
        inner_size(num): the dimensionality in feed-forward layer. Default: 256
        hidden_dropout_prob(float): probability of an element to be zeroed. Default: 0.5
        attn_dropout_prob(float): probability of an attention score to be zeroed. Default: 0.5
        hidden_act(str): activation function in feed-forward layer. Default: 'gelu'
                      candidates: 'gelu', 'relu', 'swish', 'tanh', 'sigmoid'
        layer_norm_eps(float): a value added to the denominator for numerical stability. Default: 1e-12

    """

    def __init__(
        self,
        n_layers=2,
        n_heads=2,
        hidden_size=64,
        inner_size=256,
        hidden_dropout_prob=0.5,
        attn_dropout_prob=0.5,
        hidden_act="gelu",
        layer_norm_eps=1e-12,
    ):
        super(TransformerEncoder, self).__init__()
        layer = TransformerLayer(
            n_heads,
            hidden_size,
            inner_size,
            hidden_dropout_prob,
            attn_dropout_prob,
            hidden_act,
            layer_norm_eps,
        )
        self.layer = nn.ModuleList([copy.deepcopy(layer) for _ in range(n_layers)])

    def forward(self, hidden_states, attention_mask, output_all_encoded_layers=True):
        """
        Args:
            hidden_states (torch.Tensor): the input of the TransformerEncoder
            attention_mask (torch.Tensor): the attention mask for the input hidden_states
            output_all_encoded_layers (Bool): whether output all transformer layers' output

        Returns:
            all_encoder_layers (list): if output_all_encoded_layers is True, return a list consists of all transformer
            layers' output, otherwise return a list only consists of the output of last transformer layer.

        """
        all_encoder_layers = []
        for layer_module in self.layer:
            hidden_states = layer_module(hidden_states, attention_mask)
            if output_all_encoded_layers:
                all_encoder_layers.append(hidden_states)
        if not output_all_encoded_layers:
            all_encoder_layers.append(hidden_states)
        return all_encoder_layers


class ItemToInterestAggregation(nn.Module):
    def __init__(self, seq_len, hidden_size, k_interests=5):
        super().__init__()
        self.k_interests = k_interests  # k latent interests
        self.theta = nn.Parameter(torch.randn([hidden_size, k_interests]))

    def forward(self, input_tensor):  # [B, L, d] -> [B, k, d]
        D_matrix = torch.matmul(input_tensor, self.theta)  # [B, L, k]
        D_matrix = nn.Softmax(dim=-2)(D_matrix)
        result = torch.einsum("nij, nik -> nkj", input_tensor, D_matrix)  # #[B, k, d]

        return result


class LightMultiHeadAttention(nn.Module):
    def __init__(
        self,
        n_heads,
        k_interests,
        hidden_size,
        seq_len,
        hidden_dropout_prob,
        attn_dropout_prob,
        layer_norm_eps,
    ):
        super(LightMultiHeadAttention, self).__init__()
        if hidden_size % n_heads != 0:
            raise ValueError(
                "The hidden size (%d) is not a multiple of the number of attention "
                "heads (%d)" % (hidden_size, n_heads)
            )

        self.num_attention_heads = n_heads
        self.attention_head_size = int(hidden_size / n_heads)
        self.all_head_size = self.num_attention_heads * self.attention_head_size

        # initialization for low-rank decomposed self-attention
        self.query = nn.Linear(hidden_size, self.all_head_size)
        self.key = nn.Linear(hidden_size, self.all_head_size)
        self.value = nn.Linear(hidden_size, self.all_head_size)

        self.attpooling_key = ItemToInterestAggregation(
            seq_len, hidden_size, k_interests
        )
        self.attpooling_value = ItemToInterestAggregation(
            seq_len, hidden_size, k_interests
        )

        # initialization for decoupled position encoding
        self.attn_scale_factor = 2
        self.pos_q_linear = nn.Linear(hidden_size, self.all_head_size)
        self.pos_k_linear = nn.Linear(hidden_size, self.all_head_size)
        self.pos_scaling = (
            float(self.attention_head_size * self.attn_scale_factor) ** -0.5
        )
        self.pos_ln = nn.LayerNorm(hidden_size, eps=layer_norm_eps)

        self.attn_dropout = nn.Dropout(attn_dropout_prob)

        self.dense = nn.Linear(hidden_size, hidden_size)
        self.LayerNorm = nn.LayerNorm(hidden_size, eps=layer_norm_eps)
        self.out_dropout = nn.Dropout(hidden_dropout_prob)

    def transpose_for_scores(self, x):  # transfor to multihead
        new_x_shape = x.size()[:-1] + (
            self.num_attention_heads,
            self.attention_head_size,
        )
        x = x.view(*new_x_shape)
        return x.permute(0, 2, 1, 3)

    def forward(self, input_tensor, pos_emb):
        # linear map
        mixed_query_layer = self.query(input_tensor)
        mixed_key_layer = self.key(input_tensor)
        mixed_value_layer = self.value(input_tensor)

        # low-rank decomposed self-attention: relation of items
        query_layer = self.transpose_for_scores(mixed_query_layer)
        key_layer = self.transpose_for_scores(self.attpooling_key(mixed_key_layer))
        value_layer = self.transpose_for_scores(
            self.attpooling_value(mixed_value_layer)
        )

        attention_scores = torch.matmul(query_layer, key_layer.transpose(-1, -2))
        attention_scores = attention_scores / math.sqrt(self.attention_head_size)

        # normalize the attention scores to probabilities.
        attention_probs = nn.Softmax(dim=-2)(attention_scores)
        attention_probs = self.attn_dropout(attention_probs)
        context_layer_item = torch.matmul(attention_probs, value_layer)

        # decoupled position encoding: relation of positions
        value_layer_pos = self.transpose_for_scores(mixed_value_layer)
        pos_emb = self.pos_ln(pos_emb).unsqueeze(0)
        pos_query_layer = (
            self.transpose_for_scores(self.pos_q_linear(pos_emb)) * self.pos_scaling
        )
        pos_key_layer = self.transpose_for_scores(self.pos_k_linear(pos_emb))

        abs_pos_bias = torch.matmul(pos_query_layer, pos_key_layer.transpose(-1, -2))
        abs_pos_bias = abs_pos_bias / math.sqrt(self.attention_head_size)
        abs_pos_bias = nn.Softmax(dim=-2)(abs_pos_bias)

        context_layer_pos = torch.matmul(abs_pos_bias, value_layer_pos)

        context_layer = context_layer_item + context_layer_pos

        context_layer = context_layer.permute(0, 2, 1, 3).contiguous()
        new_context_layer_shape = context_layer.size()[:-2] + (self.all_head_size,)
        context_layer = context_layer.view(*new_context_layer_shape)
        hidden_states = self.dense(context_layer)
        hidden_states = self.out_dropout(hidden_states)
        hidden_states = self.LayerNorm(hidden_states + input_tensor)

        return hidden_states


class LightTransformerLayer(nn.Module):
    """
    One transformer layer consists of a multi-head self-attention layer and a point-wise feed-forward layer.

    Args:
        hidden_states (torch.Tensor): the input of the multi-head self-attention sublayer
        attention_mask (torch.Tensor): the attention mask for the multi-head self-attention sublayer

    Returns:
        feedforward_output (torch.Tensor): the output of the point-wise feed-forward sublayer, is the output of the transformer layer
    """

    def __init__(
        self,
        n_heads,
        k_interests,
        hidden_size,
        seq_len,
        intermediate_size,
        hidden_dropout_prob,
        attn_dropout_prob,
        hidden_act,
        layer_norm_eps,
    ):
        super(LightTransformerLayer, self).__init__()
        self.multi_head_attention = LightMultiHeadAttention(
            n_heads,
            k_interests,
            hidden_size,
            seq_len,
            hidden_dropout_prob,
            attn_dropout_prob,
            layer_norm_eps,
        )
        self.feed_forward = FeedForward(
            hidden_size,
            intermediate_size,
            hidden_dropout_prob,
            hidden_act,
            layer_norm_eps,
        )

    def forward(self, hidden_states, pos_emb):
        attention_output = self.multi_head_attention(hidden_states, pos_emb)
        feedforward_output = self.feed_forward(attention_output)
        return feedforward_output


class LightTransformerEncoder(nn.Module):
    r"""One LightTransformerEncoder consists of several LightTransformerLayers.

    Args:
        n_layers(num): num of transformer layers in transformer encoder. Default: 2
        n_heads(num): num of attention heads for multi-head attention layer. Default: 2
        hidden_size(num): the input and output hidden size. Default: 64
        inner_size(num): the dimensionality in feed-forward layer. Default: 256
        hidden_dropout_prob(float): probability of an element to be zeroed. Default: 0.5
        attn_dropout_prob(float): probability of an attention score to be zeroed. Default: 0.5
        hidden_act(str): activation function in feed-forward layer. Default: 'gelu'.
            candidates: 'gelu', 'relu', 'swish', 'tanh', 'sigmoid'
        layer_norm_eps(float): a value added to the denominator for numerical stability. Default: 1e-12
    """

    def __init__(
        self,
        n_layers=2,
        n_heads=2,
        k_interests=5,
        hidden_size=64,
        seq_len=50,
        inner_size=256,
        hidden_dropout_prob=0.5,
        attn_dropout_prob=0.5,
        hidden_act="gelu",
        layer_norm_eps=1e-12,
    ):
        super(LightTransformerEncoder, self).__init__()
        layer = LightTransformerLayer(
            n_heads,
            k_interests,
            hidden_size,
            seq_len,
            inner_size,
            hidden_dropout_prob,
            attn_dropout_prob,
            hidden_act,
            layer_norm_eps,
        )
        self.layer = nn.ModuleList([copy.deepcopy(layer) for _ in range(n_layers)])

    def forward(self, hidden_states, pos_emb, output_all_encoded_layers=True):
        """
        Args:
            hidden_states (torch.Tensor): the input of the TrandformerEncoder
            attention_mask (torch.Tensor): the attention mask for the input hidden_states
            output_all_encoded_layers (Bool): whether output all transformer layers' output

        Returns:
            all_encoder_layers (list): if output_all_encoded_layers is True, return a list consists of all transformer layers' output,
            otherwise return a list only consists of the output of last transformer layer.
        """
        all_encoder_layers = []
        for layer_module in self.layer:
            hidden_states = layer_module(hidden_states, pos_emb)
            if output_all_encoded_layers:
                all_encoder_layers.append(hidden_states)
        if not output_all_encoded_layers:
            all_encoder_layers.append(hidden_states)
        return all_encoder_layers



class CNNLayers(nn.Module):
    r"""CNNLayers

    Args:
        - channels(list): a list contains the channels of each layer in cnn layers
        - kernel(list): a list contains the kernels of each layer in cnn layers
        - strides(list): a list contains the channels of each layer in cnn layers
        - activation(str): activation function after each layer in mlp layers. Default: 'relu'
                      candidates: 'sigmoid', 'tanh', 'relu', 'leekyrelu', 'none'

    Shape:
        - Input: :math:`(N, C_{in}, H_{in}, W_{in})`
        - Output: :math:`(N, C_{out}, H_{out}, W_{out})` where

        .. math::
            H_{out} = \left\lfloor\frac{H_{in}  + 2 \times \text{padding}[0] - \text{dilation}[0]
                      \times (\text{kernel\_size}[0] - 1) - 1}{\text{stride}[0]} + 1\right\rfloor

        .. math::
            W_{out} = \left\lfloor\frac{W_{in}  + 2 \times \text{padding}[1] - \text{dilation}[1]
                      \times (\text{kernel\_size}[1] - 1) - 1}{\text{stride}[1]} + 1\right\rfloor

    Examples::

        >>> m = CNNLayers([1, 32, 32], [2,2], [2,2], 'relu')
        >>> input = torch.randn(128, 1, 64, 64)
        >>> output = m(input)
        >>> print(output.size())
        >>> torch.Size([128, 32, 16, 16])
    """

    def __init__(self, channels, kernels, strides, activation="relu", init_method=None):
        super(CNNLayers, self).__init__()
        self.channels = channels
        self.kernels = kernels
        self.strides = strides
        self.activation = activation
        self.init_method = init_method
        self.num_of_nets = len(self.channels) - 1

        if len(kernels) != len(strides) or self.num_of_nets != (len(kernels)):
            raise RuntimeError("channels, kernels and strides don't match\n")

        cnn_modules = []

        for i in range(self.num_of_nets):
            cnn_modules.append(
                nn.Conv2d(
                    self.channels[i],
                    self.channels[i + 1],
                    self.kernels[i],
                    stride=self.strides[i],
                )
            )
            if self.activation.lower() == "sigmoid":
                cnn_modules.append(nn.Sigmoid())
            elif self.activation.lower() == "tanh":
                cnn_modules.append(nn.Tanh())
            elif self.activation.lower() == "relu":
                cnn_modules.append(nn.ReLU())
            elif self.activation.lower() == "leakyrelu":
                cnn_modules.append(nn.LeakyReLU())
            elif self.activation.lower() == "none":
                pass

        self.cnn_layers = nn.Sequential(*cnn_modules)

        if self.init_method is not None:
            self.apply(self.init_weights)

    def init_weights(self, module):
        # We just initialize the module with normal distribution as the paper said
        if isinstance(module, nn.Conv2d):
            if self.init_method == "norm":
                normal_(module.weight.data, 0, 0.01)
            if module.bias is not None:
                module.bias.data.fill_(0.0)

    def forward(self, input_feature):
        return self.cnn_layers(input_feature)



class SparseDropout(nn.Module):
    """
    This is a Module that execute Dropout on Pytorch sparse tensor.
    """

    def __init__(self, p=0.5):
        super(SparseDropout, self).__init__()
        # p is ratio of dropout
        # convert to keep probability
        self.kprob = 1 - p

    def forward(self, x):
        if not self.training:
            return x

        mask = ((torch.rand(x._values().size()) + self.kprob).floor()).type(torch.bool)
        rc = x._indices()[:, mask]
        val = x._values()[mask] * (1.0 / self.kprob)
        return torch.sparse.FloatTensor(rc, val, x.shape)


import os
import csv
import math
import gzip
import pickle
import pandas as pd
import numpy as np
from tqdm import tqdm
from sklearn.preprocessing import StandardScaler


def split_list(origin_list, n):
    """
    Splits the input list into smaller sublists of size n (or close to n).

    :param origin_list: The original list to be split.
    :param n: The number of sublists to split into.
    :return: A list of sublists.
    """
    res_list = []
    L = len(origin_list)
    N = int(math.ceil(L / float(n)))
    begin = 0
    end = begin + N
    while begin < L:
        if end < L:
            temp_list = origin_list[begin:end]
            res_list.append(temp_list)
            begin = end
            end += N
        else:
            temp_list = origin_list[begin:]
            res_list.append(temp_list)
            break
    return res_list


def load_embedding(filename, sep = '\t'):
    """
    Load embedding from file
    :param filename: embedding file name
    :param sep: the char used as separation symbol
    :return: a dict with item name as key and embedding vector as value
    """

    with open(filename, 'r') as fp:
        result = {}
        for l in fp:
            l = l.strip()
            if l == '':
                continue
            sp = l.split(sep)
            vals = [float(sp[i]) for i in range(1, len(sp))]
            result[sp[0]] = vals
        return result
    

def get_rel_feat(path):
    """
    Loads and scales the relevance features from a CSV file.

    :param path: Path to the CSV file containing the relevance features.
    :return: A dictionary where the key is a tuple (query, doc) and the value is a list of features.
    """
    rel_feat = pd.read_csv(path)
    rel_feat_names = list(sorted(set(rel_feat.columns) - {'query', 'doc'}))
    rel_feat[rel_feat_names] = StandardScaler().fit_transform(rel_feat[rel_feat_names])
    rel_feat = dict(zip(map(lambda x: tuple(x), rel_feat[['query', 'doc']].values),
            rel_feat[rel_feat_names].values.tolist()))
    return rel_feat


def read_rel_feat(path):
    """
    Reads relevance features from a CSV file and returns them in a nested dictionary format.

    :param path: Path to the CSV file containing the relevance features.
    :return: A nested dictionary where the key is a query and the value is another dictionary of documents and features.
    """
    rel_feat = {}
    f = csv.reader(open(path, 'r'), delimiter = ',')
    next(f)
    for line in f:
        if line[0] not in rel_feat:
            rel_feat[line[0]] = {}
        if line[1] not in rel_feat[line[0]]:
            rel_feat[line[0]][line[1]] = np.array([float(val) for val in line[2:]])
    return rel_feat


def pkl_load(filename):
    """
    Loads a pickle file and returns the data inside it.

    :param filename: Path to the pickle file.
    :return: The loaded data from the pickle file.
    """
    if not os.path.exists(filename):
        print('filename={} not exists!')
        return
    with gzip.open(filename, 'rb') as f:
        data_dict = pickle.load(f)
    return data_dict


def pkl_save(data_dict, filename):
    """ Saves a dictionary to a compressed pickle file.

    :param data_dict: The dictionary to be saved.
    :param filename: The path where the pickle file should be saved.
    """
    with gzip.open(filename, 'wb') as f:
        pickle.dump(data_dict, f)


def remove_duplicate(input_path, output_path):
    """ Removes duplicate documents in the ranking list.

    :param input_path: The path to the input file containing the ranking list.
    :param output_path: The path where the cleaned ranking list will be saved.
    """
    unique_records = set()
    output_lines = []

    with open(input_path, "r") as file:
        for line in file:
            parts = line.strip().split()
            topic_id = parts[0]
            document_name = parts[2]
            
            key = (topic_id, document_name)
            
            if key not in unique_records:
                unique_records.add(key)
                output_lines.append(line)

    with open(output_path, "w") as file:
        for line in output_lines:
            file.write(line)


def restore_doc_ids(order_str, id_dict):
    """ Restores document IDs based on an ordered list of indices and a dictionary of document IDs.

    :param order_str: A string representing the order of document indices.
    :param id_dict: A dictionary mapping indices to document IDs.
    :return: A list of document IDs in the restored order.
    """
    order = [int(x) for x in order_str.replace(" ", "").replace("[", "").replace("]", "").split(">")]
    reversed_dict = {v: k for k, v in id_dict.items()}
    return [reversed_dict[num] for num in order if num in reversed_dict]


def get_metrics_20(csv_file_path):
    """
    Retrieves evaluation metrics from a CSV file for the top 20 documents.

    :param csv_file_path: The path to the CSV file containing evaluation results.
    :return: A tuple containing the mean values of alpha-nDCG@20, NRBP@20, ERR-IA@20, and strec@20.
    """
    all_qids=range(1,201)
    del_index=[94,99]
    all_qids=np.delete(all_qids,del_index)
    qids=[str(i) for i in all_qids]

    df=pd.read_csv(csv_file_path)

    alpha_nDCG_20=df.loc[df['topic'].isin(qids)]['alpha-nDCG@20'].mean()
    NRBP_20=df.loc[df['topic'].isin(qids)]['NRBP'].mean()
    ERR_IA_20=df.loc[df['topic'].isin(qids)]['ERR-IA@20'].mean()
    # Pre_IA_20=df.loc[df['topic'].isin(qids)]['P-IA@20'].mean()
    S_rec_20=df.loc[df['topic'].isin(qids)]['strec@20'].mean()
    
    
    return alpha_nDCG_20, NRBP_20, ERR_IA_20, S_rec_20




import json
import requests
from requests.adapters import HTTPAdapter
from urllib3.util import Retry
import time
import re


class ApiException(Exception):
    """
    Custom exception class for handling API-specific errors.
    
    :param msg: Error message describing the API failure
    :param error_code: HTTP status code from the API response
    """

    def __init__(self, msg, error_code):
        self.msg = msg
        self.error_code = error_code


class ApiProxy():
    """
    A proxy class for handling HTTP communications with LLM API endpoints.
    This class manages API request sessions, implements retry logic, and handles
    various HTTP status codes with exponential backoff for rate limits and server errors.
    
    :param url: The API endpoint URL
    :param api_key: Optional API key for authentication with LLM services
    """

    def __init__(self, url, api_key=None):
        retry_strategy = Retry(
            total=1,
            backoff_factor=1,
            status_forcelist=[429, 500, 502, 503, 504],
            allowed_methods=["POST"],
        )
        adapter = HTTPAdapter()
        self.session = requests.Session()
        self.session.mount("https://", adapter)
        self.session.mount("http://", adapter)
        self.api_key = api_key
        self.url = url

    def __call__(self, params_gpt, headers={}):
        headers['Content-Type'] = headers['Content-Type'] if 'Content-Type' in headers else 'application/json'
        if self.api_key:
            headers['Authorization'] = "Bearer " + self.api_key
        try:
            response = self.session.post(self.url, headers=headers, data=json.dumps(params_gpt))
        except Exception as e:
            time.sleep(10)
            response = self.session.post(self.url, headers=headers, data=json.dumps(params_gpt))

        if response.status_code in (429, 404, 500, 502, 503, 504, 104):
            for idx in range(10000000000000000):
                interval = min(1, 10+idx*5)
                if idx % 10 == 0:
                    print("meet error 429 or 404, code: {}, msg: {},run {}".format(
                        response.status_code, response.text, interval, idx
                    ), flush=True)

                response = self.session.post(self.url, headers=headers, data=json.dumps(params_gpt))
                if response.status_code not in (429, 404, 500, 502, 503, 504, 104):
                    break
        if response.status_code != 200:
            err_msg = f"access error, status code: {response.status_code}, errmsg: {response.text}"
            raise ApiException(err_msg, response.status_code)
        data = json.loads(response.text)
        return data


class LMAgent():
    """
    A high-level interface for interacting with large language models.
    
    :param config: Dictionary containing model configuration parameters
    """

    def __init__(self, config):
        self.model = config["model_name"]
        self.apikey = config["api_key"]
        self.temperature = config["temperature"]
        self.max_new_tokens = config["max_new_tokens"]
        self.top_p = config["top_p"]
        self._proxy = ApiProxy(url=config['api_url']+'/v1/chat/completions', api_key=self.apikey)

    def __call__(self, system_prompt, input_prompt, max_new_tokens):
        if isinstance(input_prompt, list):
            system_message, input_message = input_prompt[0].split("## Input Data")
        else:
            system_message, input_message = input_prompt.split("## Input Data")
        messages = [
            {"role": "system", "content": system_message},
            {"role": "user", "content": "## Input Data"+input_message}
        ]

        prompt_dict = {
            "model": self.model,
            "messages": messages,
            "temperature": self.temperature,
            "max_tokens": self.max_new_tokens,
            "top_p": self.top_p,
            "frequency_penalty": 0.0,
            "presence_penalty": 0.0,
        }
        response = self._proxy(prompt_dict)
        content = response["choices"][0]["message"]["content"]

        pattern = r'\{"rerank_list": "[^"]+"\}'

        match = re.search(pattern, content, re.DOTALL)
        if match:
            # Convert the string to a dictionary
            result_dict = match.group()
            return [result_dict] if isinstance(input_prompt, list) else result_dict
        else:
            return [content] if isinstance(input_prompt, list) else content
    



import numpy as np
#from
import os

import pandas as pd
import yaml
import random

from .process_dataset import Process
from .base_model import MF, GRU4Rec, SASRec, BPR, BPR_Seq
from .rank_model import IPS, SDRO, Minmax_SGD, APR, FOCF, FairDual, Reg, FairNeg, DPR

import time
import torch.optim as optim

from .sampler import PointWiseDataset, PairWiseDataset, RankingTestDataset, SequentialDataset
from .evaluator import CTR_Evaluator, Ranking_Evaluator

from torch.utils.data import Dataset, DataLoader
from tqdm import tqdm,trange
from datetime import datetime
import torch
import json
from scipy.sparse import save_npz, load_npz
import torch.nn as nn
import ast



class RecTrainer(object):
    def __init__(self, train_config):
        """Initialize In-processing and base models.

        :param train_config: Your custom config files.
        """

        self.dataset = train_config['dataset']
        #self.stage = stage
        self.train_config = train_config


    def load_configs(self, dir):
        """
            Loads and merges configuration files for the model, dataset, and evaluation.

            This function loads multiple YAML configuration files, including the process configuration,
            dataset-specific settings, model configurations, and evaluation parameters. All configurations
            are merged, with the highest priority given to the class's own `config` attribute.

            :param dir: The directory where the main process configuration file is located.
            :return: A dictionary containing the merged configuration from all files.
        """


        print("start to load config...")
        with open(os.path.join(dir, "process_config.yaml"), 'r') as f:
            config = yaml.safe_load(f)


        # print(train_data_df.head())

        print("start to load model...")
        with open(os.path.join("recommendation", "properties", "models.yaml"), 'r') as f:
            model_config = yaml.safe_load(f)

        with open(os.path.join("recommendation", "properties", "models", self.train_config['model'] + ".yaml"),
                  'r') as f:
            model_config.update(yaml.safe_load(f))

        if self.train_config['fair-rank'] == True:
            with open(os.path.join("recommendation", "properties", "models", self.train_config['rank_model'] + ".yaml"),
                      'r') as f:
                model_config.update(yaml.safe_load(f))

        config.update(model_config)

        with open(os.path.join("recommendation", "properties", "evaluation.yaml"), 'r') as f:
            config.update(yaml.safe_load(f))

        config.update(self.train_config)  ###train_config has highest rights
        print("your loading config is:")
        print(config)

        return config

    def Set_Dataset(self, data_type, config, train_data_df, val_data_df, test_data_df):
        """
            Initializes and returns the training, validation, and test datasets based on the specified data type and evaluation type.

            This function creates appropriate dataset objects for training, validation, and testing based on the provided data type
            (point, pair, or sequential) and the evaluation type (CTR or ranking). It supports different dataset types for training
            and evaluation, and raises an error if an unsupported type is provided.

            :param data_type: The type of dataset to be used for training. Must be one of ['point', 'pair', 'sequential'].
            :param config: A configuration dictionary that contains parameters for dataset creation and evaluation type.
            :param train_data_df: The DataFrame containing the training data.
            :param val_data_df: The DataFrame containing the validation data.
            :param test_data_df: The DataFrame containing the test data.
            :return: A tuple containing the training, validation, and test datasets.
        """
        if data_type == 'point':
            train = PointWiseDataset(train_data_df, config)
        elif data_type == 'pair':
            train = PairWiseDataset(train_data_df, config)
        elif data_type == 'sequential':
            train = SequentialDataset(train_data_df, config)
        else:
            raise NotImplementedError("train_type only supports in [point, pair, sequential]")

        if config['eval_type'] == 'CTR':
            valid = PointWiseDataset(val_data_df, config)
            test = PointWiseDataset(test_data_df, config)
        elif config['eval_type'] == 'ranking':
            valid = RankingTestDataset(val_data_df, config)
            test = RankingTestDataset(test_data_df, config)
        else:
            raise NotImplementedError("We only support the eval type as [CTR, ranking]")

        return train, valid, test

    def check_model_stage(self,config, Model):
        """
            Checks if the provided data type in the configuration aligns with the supported model type.

            This function verifies that the data type specified in the `config` dictionary is compatible with the model's
            supported types. If the data type is not supported by the model, a `ValueError` is raised with an informative message.

            :param config: A configuration dictionary that includes the data type used for testing.
            :param Model: The model class or object which has a `type` attribute specifying the supported data types.
        """
        if config['data_type'] not in Model.type:
            raise ValueError(f"The tested data type does not align with the model type: input is {config['data_type']}, "
                             f"the model only support: {Model.type}")

    def train(self):
        """
            Training in-processing and base model main workflow.
        """

        dir = os.path.join("recommendation", "processed_dataset", self.dataset)

        state = Process(self.dataset, self.train_config)
        config = self.load_configs(dir)
        print(state)
        #exit(0)

        print("start to load dataset......")

        self.device = config['device']

        if config['model'] == 'mf':
            self.Model = MF(config).to(self.device)
        elif config['model'] == 'BPR':
            self.Model = BPR(config).to(self.device)
        elif config['model'] == 'BPR_Seq':
            self.Model = BPR_Seq(config).to(self.device)
        elif config['model'] == 'gru4rec':
            self.Model = GRU4Rec(config).to(self.device)
        elif config['model'] == 'SASRec':
            self.Model = SASRec(config).to(self.device)

        else:
            raise NotImplementedError(f"Not supported model type: {config['model']}")

        self.check_model_stage(config, self.Model)

        self.group_weight = np.ones(config['group_num'])





        train_data_df = pd.read_csv(os.path.join(dir, self.dataset + ".train"), sep='\t')
        val_data_df = pd.read_csv(os.path.join(dir, self.dataset + ".valid." + config['eval_type']), sep='\t')
        test_data_df = pd.read_csv(os.path.join(dir, self.dataset + ".test." + config['eval_type']), sep='\t')

        print(test_data_df.head())
        train_data_df["history_behaviors"] = train_data_df["history_behaviors"].apply(lambda x: np.array(ast.literal_eval(x)))
        val_data_df["history_behaviors"] = val_data_df["history_behaviors"].apply(lambda x: np.array(ast.literal_eval(x)))
        test_data_df["history_behaviors"] = test_data_df["history_behaviors"].apply(lambda x: np.array(ast.literal_eval(x)))

        optimizer = optim.Adam(self.Model.parameters(), lr= config['learning_rate'])
        data_type = config['data_type']



        train, valid, test = self.Set_Dataset(data_type, config, train_data_df, val_data_df, test_data_df)

        if config['fair-rank'] == True:
            if config['data_type'] == 'point':
                raise ValueError(
                    "fair ranking model only supports the pair and sequential data_type, not the point type")

            if config['rank_model'] == "IPS":
                self.Fair_Ranker = IPS(config, self.group_weight)

            elif config['rank_model'] == 'SDRO':
                self.Fair_Ranker = SDRO(config, self.group_weight)

            elif config['rank_model'] == 'Minmax_SGD':
                self.Fair_Ranker = Minmax_SGD(config, self.group_weight)

            elif config['rank_model'] == 'APR':
                self.Fair_Ranker = APR(config, self.group_weight)

            elif config['rank_model'] == 'FOCF':
                self.Fair_Ranker = FOCF(config, self.group_weight)

            elif config['rank_model'] == 'FairDual':
                self.Fair_Ranker = FairDual(config, self.group_weight)

            elif config['rank_model'] == 'Reg':
                self.Fair_Ranker = Reg(config, self.group_weight)

            elif config['rank_model'] == 'FairNeg':
                self.Fair_Ranker = FairNeg(config, train.user2pos)

            elif config['rank_model'] == 'DPR':
                self.Fair_Ranker = DPR(config, self.group_weight)

            else:
                NotImplementedError(f"Not supported fair rank model type:{config['rank_model']}")

            self.check_model_stage(config, self.Fair_Ranker)
            if self.Fair_Ranker.fair_type == "sample" and config['data_type'] != "pair":
                raise ValueError(
                    f"The choosed fair ranker [{config['rank_model']}] only support the base model type as pair")


        train_loader = DataLoader(train, batch_size=config['batch_size'], shuffle=True)
        valid_loader = DataLoader(valid, batch_size=config['eval_batch_size'], shuffle=False)
        test_loader = DataLoader(test, batch_size=config['eval_batch_size'], shuffle=False)

        if config['eval_type'] == 'CTR':
            evaluator = CTR_Evaluator(config)
        elif config['eval_type'] == 'ranking':
            evaluator = Ranking_Evaluator(config)
        else:
            raise NotImplementedError("we only support eval type in [CTR, ranking] !")

        today = datetime.today()
        today_str = f"{today.year}-{today.month}-{today.day}"
        log_dir = os.path.join("recommendation", "log", f"{today_str}_{config['log_name']}")
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)

        print("start to train...")

        for epoch in trange(config['epoch']):


            total_loss = 0
            self.Model.train()
            best_result = -1

            if config['fair-rank'] == True:
                if 'update_epoch' in config and epoch % config['update_epoch'] == 0:
                    if config['rank_model'] == 'FairDual':
                        self.Fair_Ranker.reset_parameters(len(train_data_df))
                    else:
                        self.Fair_Ranker.reset_parameters()

            #for user_ids, item_ids, group_ids, label in train_loader:
            for train_datas in train_loader:

                if data_type == "point":
                    interaction = {"user_ids": train_datas[0].to(self.device), "history_ids": train_datas[1].to(self.device),
                                   "item_ids": train_datas[2].to(self.device),
                                   "group_ids": train_datas[3].to(self.device), "label": train_datas[4].to(self.device)}
                    #ids = {"user_ids": train_datas[0].to(self.device), "item_ids": train_datas[1].to(self.device)}
                elif data_type == "pair":
                    interaction = {"user_ids": train_datas[0].to(self.device), "history_ids": train_datas[1].to(self.device),
                                   "item_ids": train_datas[2].to(self.device), "neg_item_ids": train_datas[3].to(self.device),
                                   "group_ids": train_datas[4].to(self.device), "neg_group_ids": train_datas[5].to(self.device)
                                   }
                    #ids = {"user_ids": train_datas[0].to(self.device), "item_ids": train_datas[1].to(self.device)}
                else: ###squential format
                    interaction = {"user_ids": train_datas[0].to(self.device), "history_ids": train_datas[1].to(self.device),
                                   "item_ids": train_datas[2].to(self.device), "group_ids": train_datas[3].to(self.device), }

                feed_user_dict = {"user_ids": train_datas[0].to(self.device), "history_ids": train_datas[1].to(self.device)}
                feed_item_ids = train_datas[2].to(self.device)

                optimizer.zero_grad()
                #loss = self.Model.compute_loss(interaction)

                if config['fair-rank'] == True:

                    if self.Fair_Ranker.fair_type == 're-weight':
                        loss = self.Model.compute_loss(interaction)
                        scores = self.Model(feed_user_dict, feed_item_ids)


                        input_dict = {'items': train_datas[2].detach().numpy(), 'loss': loss.detach().cpu().numpy(),
                                      'scores': scores.detach().cpu().numpy()}

                        if config['rank_model'] == 'FairDual':
                            ## FairDual needs sample some items
                            exposure_sample_num = config['exposure_sample_num']
                            item_sample_ids = random.sample(range(config['item_num']), exposure_sample_num)
                            item_sample_ids = torch.tensor(item_sample_ids, dtype=torch.long).to(self.device)

                            scores, indices = self.Model.full_ranking(feed_user_dict, item_sample_ids, k=config['s_k'])

                            items = item_sample_ids[indices.detach().cpu().numpy()]
                            input_dict['sample_items'] = items

                        weight = self.Fair_Ranker.reweight(input_dict=input_dict)
                        #print(weight)
                        loss = torch.mean(torch.tensor(weight).to(self.device)*loss)

                    elif self.Fair_Ranker.fair_type == "sample":
                        neg_item_ids, adj = self.Fair_Ranker.sample(interaction, self.Model)
                        interaction['neg_item_ids'] = neg_item_ids
                        loss = self.Model.compute_loss(interaction)
                        if config['rank_model'] == "FairNeg":
                            group_loss = loss.detach().cpu().numpy()
                            group_loss = np.matmul(group_loss, adj)/(np.sum(adj, axis=0, keepdims=False)+1)
                            self.Fair_Ranker.accumulate_epoch_loss(group_loss)


                        loss = torch.mean(loss)

                    elif self.Fair_Ranker.fair_type == "regularizer":
                        loss = self.Model.compute_loss(interaction)
                        scores = self.Model(feed_user_dict, feed_item_ids)
                        input_dict = {'items': train_datas[2], 'loss': loss, 'scores':scores}

                        if config['rank_model'] == 'DPR':
                            input_dict = {'item_ids': train_datas[2].to(self.device), 'scores': scores,
                                          "neg_item_ids": train_datas[3].to(self.device),
                                          "group_ids": train_datas[4].to(self.device),
                                          "neg_group_ids": train_datas[5].to(self.device)
                                          }
                            fair_loss = self.Fair_Ranker.fairness_loss(input_dict, self.Model)
                            loss = torch.mean(loss) + fair_loss

                        else:
                            fair_loss = self.Fair_Ranker.fairness_loss(input_dict)
                            loss = torch.mean(loss) + config['fair_lambda'] * fair_loss

                    else: ##other type of fairness method, which modifies embeddings during the evaluation process
                        loss = self.Model.compute_loss(interaction)
                        loss = torch.mean(loss)

                else:
                    loss = self.Model.compute_loss(interaction)
                    loss = torch.mean(loss)

                loss.backward()
                optimizer.step()
                total_loss += loss.item()



            if epoch % config['eval_step'] == 0:
                eval_result = evaluator.eval(valid_loader, self.Model)
                watch_eval_value = eval_result[config['watch_metric']]
                if watch_eval_value >= best_result:
                    best_result = watch_eval_value
                    torch.save(self.Model.state_dict(), os.path.join(log_dir, "best_model.pth"))
                print(f"eval result: {eval_result}, best result: {best_result}")
                print()


            print("epoch: %d loss: %.3f" %(epoch, total_loss/ len(train_loader)))



        print(f"training complete! start to save the config and model...")
        print(f" config files are dump in {log_dir}")
        with open(os.path.join(log_dir, "config.yaml"), 'w') as f:
            yaml.dump(config, f)

        print("start to testing...")
        self.Model.load_state_dict(torch.load(os.path.join(log_dir, "best_model.pth")))  # load state_dict
        self.Model.eval()  # change to eval model
        if config['store_scores'] == False:
            test_result = evaluator.eval(test_loader, self.Model)
        else:
            test_result, coo_matrix = evaluator.eval(test_loader, self.Model, store_scores=True)
            save_npz(os.path.join(log_dir, 'ranking_scores.npz'), coo_matrix) ##prepared for re-ranking stage


        with open(os.path.join(log_dir, 'test_result.json'), 'w') as file:
            json.dump(test_result, file)
        print(test_result)
        print(f"dump in {log_dir}")





from .CPFair import CPFair
from .FairRec import FairRec
from .FairRecPlus import FairRecPlus
from .k_neighbor import k_neighbor
from .min_regularizer import min_regularizer
from .PMMF import PMMF
from .Welf import Welf
from .TaxRank import TaxRank
from .FairSync import FairSync
from .RAIF import RAIF


import os
import json
import pickle

from ..utils.utils import pkl_load, get_metrics_20
from .base import BasePostProcessUnsupervisedModel

'''
@inproceedings{santos2010exploiting,
  title={Exploiting query reformulations for web search result diversification},
  author={Santos, Rodrygo LT and Macdonald, Craig and Ounis, Iadh},
  booktitle={Proceedings of the 19th international conference on World wide web},
  pages={881--890},
  year={2010}
}
'''

class xQuAD(BasePostProcessUnsupervisedModel):
    def __init__(self, top_k=20):
        """ 
        Initializes the xQuAD for diversified document ranking.

        :param top_k: The maximum number of documents to be ranked. Default is 20.
        """
        super().__init__(top_k)
    
    def rerank(self, config):
        """ 
        Re-ranks documents using the xQuAD algorithm based on BM25 scores and query diversification.

        :param config: A dictionary containing configuration parameters.
        """
        qd = pickle.load(open(os.path.join(config['data_dir'], 'div_query.data'), 'rb'))
        bm25_dict = pkl_load(os.path.join(config['data_dir'], config['model'], 'bm25_scores.pkl'))

        output_dir = os.path.join(config['tmp_dir'], config['model'])
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        output_txt = os.path.join(output_dir, 'output_best.txt')
        csv_path = os.path.join(output_dir, 'result.csv')

        for qid in qd:
            doc_ranking = self.calculate_xquad_score(qid, bm25_dict, config['lambda'])

            judge_f = open(os.path.join(output_dir, str(qid)+'.txt'), 'w')
            for i in range(len(doc_ranking)):
                doc_ranking_score = len(doc_ranking) - i
                judge_f.write(str(qid) + ' Q0 ' + doc_ranking[i] + ' ' + str(i + 1) + ' ' + str(doc_ranking_score) + ' indri\n')
            judge_f.close()

        command = 'cat '+output_dir+'/* > '+output_txt
        os.system(command)
        command = './search/eval/clueweb09/ndeval ./search/eval/clueweb09/2009-2012.diversity.ndeval.qrels '+output_txt+' >' + str(csv_path)
        os.system(command)

        alpha_nDCG_20, NRBP_20, ERR_IA_20, S_rec_20 = get_metrics_20(csv_path)
        print('alpha_nDCG@20_std = {}, NRBP_20 = {}, ERR_IA_20 = {}, S_rec_20 = {}'.format(alpha_nDCG_20, NRBP_20, ERR_IA_20, S_rec_20))


    def calculate_xquad_score(self, qid, bm25_dict, _lambda=0.5):
        """ 
        Implements the xQuAD algorithm for diversified document ranking.

        :param qid: Query ID for which documents are being ranked.
        :param bm25_dict: Dictionary containing BM25 scores for both original and suggestion queries.
        :param _lambda: A trade-off parameter (0 to 1) that balances relevance and diversity. 
                        A higher value favors relevance, while a lower value emphasizes diversity.
        :return: A list of selected document IDs in diversified ranking order.
        """
        original_scores = bm25_dict[qid][0]  # Original query scores
        suggestion_scores = bm25_dict[qid][1]  # Suggestion query scores (n_suggestions x n_docs)
        
        selected_docs = []
        available_docs = set(original_scores.keys())
        
        # Calculate suggestion probabilities (assuming normalized BM25 scores)
        total_suggestion_score = suggestion_scores.sum(axis=0)
        suggestion_probs = suggestion_scores / (total_suggestion_score + 1e-10)  # Add epsilon to avoid division by zero
        
        while len(selected_docs) < self.top_k and available_docs:
            best_score = float('-inf')
            best_doc = None
            
            for doc_id in available_docs:
                # Get original query score component
                original_component = (1 - _lambda) * original_scores[doc_id]
                
                # Calculate expanded component
                expanded_component = 0
                doc_idx = list(original_scores.keys()).index(doc_id)
                
                for qi_idx in range(len(suggestion_scores)):
                    qi_score = suggestion_probs[qi_idx][doc_idx]
                    
                    # Calculate product term for already selected documents
                    product_term = 1
                    for selected_doc in selected_docs:
                        selected_idx = list(original_scores.keys()).index(selected_doc)
                        product_term *= (1 - suggestion_scores[qi_idx][selected_idx])
                    
                    expanded_component += qi_score * product_term
                
                expanded_component *= _lambda
                
                # Combined score
                total_score = original_component + expanded_component
                
                if total_score > best_score:
                    best_score = total_score
                    best_doc = doc_id
            
            if best_doc is None:
                break
                
            selected_docs.append(best_doc)
            available_docs.remove(best_doc)
        
        return selected_docs





"""Learning fair representations is a pre-processing technique that finds a
    latent representation which encodes the data well but obfuscates information
    about protected attributes [2]_.
    References:
        .. [2] R. Zemel, Y. Wu, K. Swersky, T. Pitassi, and C. Dwork,  "Learning
           Fair Representations." International Conference on Machine Learning,
           2013.
    Based on code from https://github.com/zjelveh/learning-fair-representations
    """

import numpy as np
import os
from ..probabilistic_mapping_helpers import get_xhat_y_hat
def compute_reconstruction_loss(x_orig, x_hat):
    """Computes the reconstruction loss (L_x)."""

    return np.mean((x_hat - x_orig) * (x_hat - x_orig))

def compute_LFR_utility_loss(y_true, y_hat):
    """Computes the utility loss (L_y) using binary cross-entropy between positive and negative samples."""
    return -np.mean(y_true * np.log(y_hat) + (1. - y_true) * np.log(1. - y_hat))


def compute_LFR_fairness_loss_pairwise(M_unpriv, M_priv):
    """Computes the fairness loss (L_z)."""
    group_gap = np.mean(abs(np.mean(M_unpriv, axis=0) - np.mean(M_priv, axis=0)))

    return group_gap
def compute_LFR_group_fairness_loss(X, params, k, group_weights, sensitive_groups, sensitive_indexes, biggest_gap):
    """LFR optimisation function"""
    L_z = 0
    for s_attr in sensitive_groups["privileged_groups"].keys():
        privileged_groups = sensitive_groups["privileged_groups"][s_attr]
        unprivileged_groups = sensitive_groups["unprivileged_groups"][s_attr]
        for index_advantaged_group in range(len(privileged_groups)):
            for s_index in sensitive_indexes:
                mask_privileged = np.isin(element=X[:, s_index],
                                          test_elements=privileged_groups[index_advantaged_group]).reshape(-1)
                if sum(mask_privileged) != 0:
                    for index_unprivileged_group in range(index_advantaged_group + 1, len(unprivileged_groups)):
                        if unprivileged_groups[index_unprivileged_group] != privileged_groups[index_advantaged_group]:
                            mask_unprivileged = np.isin(element=X[:, s_index],
                                             test_elements=unprivileged_groups[index_unprivileged_group]).reshape(-1)
                            if sum(mask_unprivileged) != 0:
                                if unprivileged_groups[index_unprivileged_group] != privileged_groups[
                                    index_advantaged_group]:
                                    weight = abs(
                                        group_weights[privileged_groups[index_advantaged_group]] - group_weights[
                                            unprivileged_groups[index_unprivileged_group]]) + 1

                                    M_unpriv, X_hat, Y_hat = get_xhat_y_hat(X[mask_unprivileged], params, k)
                                    M_priv, X_hat, Y_hat = get_xhat_y_hat(X[mask_privileged], params, k)
                                    group_gap = compute_LFR_fairness_loss_pairwise(M_unpriv, M_priv)
                                    if biggest_gap:
                                        if biggest_gap_groups < group_gap:
                                            biggest_gap_groups = group_gap
                                            L_z = weight * group_gap
                                    else:
                                        L_z += weight * group_gap

    return L_z


def LFR_optimisation(
        params, X, Y, sensitive_groups, sensitive_indexes, k=10,
        A_x=0.01, A_y=0.1, A_z=0.5,
        group_weights=None, biggest_gap=False, logs_path=''):
    """
    Objective function for the Learned Fair Representations (LFR) optimization.
    """

    M, X_hat, Y_hat = get_xhat_y_hat(X, params, k)

    L_x = compute_reconstruction_loss(X, X_hat)
    L_y = compute_LFR_utility_loss(Y, Y_hat)
    L_z = compute_LFR_group_fairness_loss(X, params, k, group_weights, sensitive_groups, sensitive_indexes,
                                          biggest_gap)

    # Compute total loss
    total_loss = A_x * L_x + A_y * L_y + A_z * L_z

    # Logging
    if LFR_optimisation.steps % 100 == 0:
        log_msg = f"step: {LFR_optimisation.steps}, loss: {total_loss}, " \
                  f"L_x: {L_x}, L_y: {L_y}, L_z: {L_z}"
        print(log_msg)

        if logs_path:
            os.makedirs(logs_path, exist_ok=True)
            with open(os.path.join(logs_path, "logs.txt"), 'a') as f:
                f.write(log_msg + "\n")

    LFR_optimisation.steps += 1
    return total_loss
